import os
import ast
import sys
import json
import argparse


class AutoGenMapper(ast.NodeVisitor):
    
    def __init__(self, filename: str | None = None):
        self.current_filename = filename
        self.agents = {}
        self.tools = {}
        self.functions = {}
        self.tool_calls = {}
        self.agent_tool_func_map = {}

        self.known_agent_bases = {"Agent", "RoutedAgent"}
        self.known_tool_wrappers = {"Tool", "FunctionTool"}


    def visit_Import(self, node: ast.Import):
        """
        Visit an Import node and extract relevant information.
        """
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        """
        Visit an Import node and extract relevant information.
        """
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign):
        if isinstance(node.value, ast.List):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    tool_list = []
                    for elt in node.value.elts:
                        if isinstance(elt, ast.Call):
                            if (
                                isinstance(elt.func, ast.Name)
                                and elt.func.id in self.known_tool_wrappers
                            ):
                                if elt.args:
                                    func_node = elt.args[0]
                                    if isinstance(func_node, ast.Name):
                                        tool_list.append(func_node.id)

                    if tool_list:
                        print(f"Found tool list: {target.id} → {tool_list}")
                        self.tools[target.id] = tool_list

        self.generic_visit(node)


    def visit_ClassDef(self, node: ast.ClassDef):
        base_names = [
            base.id for base in node.bases 
            if isinstance(base, ast.Name) and base.id in self.known_agent_bases
        ]

        if base_names:
            agent_info = {
                "type": "agent",
                "tools": [],
                "inherits": base_names
            }

            for item in node.body:
                if isinstance(item, ast.FunctionDef) and item.name == "__init__":
                    for stmt in item.body:
                        if isinstance(stmt, ast.Assign):
                            for target in stmt.targets:
                                if (isinstance(target, ast.Attribute) and target.attr == "_tools"):
                                    if isinstance(stmt.value, ast.Name):
                                        agent_info["tools"].append(stmt.value.id)

            self.agents[node.name] = agent_info

        self.generic_visit(node)


    def visit_FunctionDef(self, node: ast.FunctionDef):
        self._register_function_metadata(node)
        self._track_tool_calls(node)
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef):
        self._register_function_metadata(node)
        self._track_tool_calls(node)
        self.generic_visit(node)


    def _register_function_metadata(self, node):
        self.functions[node.name] = {
            "docstring": ast.get_docstring(node),
            "source_location": {
                "file": self.current_filename,
                "line": node.lineno,
                "col": node.col_offset,
                "end_line": getattr(node, "end_lineno", node.lineno),
                "end_col": getattr(node, "end_col_offset", node.col_offset)
            }
        }

    def _track_tool_calls(self, node: ast.FunctionDef | ast.AsyncFunctionDef):
        current_func = node.name
        called_funcs = set()

        class CallVisitor(ast.NodeVisitor):
            def visit_Call(self, call_node: ast.Call):
                if isinstance(call_node.func, ast.Name):
                    called_funcs.add(call_node.func.id)
                self.generic_visit(call_node)

        CallVisitor().visit(node)

        if called_funcs:
            self.tool_calls[str(current_func)] = list(called_funcs)


    def visit_Await(self, node: ast.Await):
        # Look for calls like: await SomeAgent.register(...)
        if isinstance(node.value, ast.Call):
            call = node.value
            if isinstance(call.func, ast.Attribute) and call.func.attr == "register":
                if len(call.args) >= 3:
                    agent_class = self._get_id(call.func.value)
                    agent_name = self._get_str(call.args[1])
                    lambda_fn = call.args[2]
                    
                    if isinstance(lambda_fn, ast.Lambda):
                        if isinstance(lambda_fn.body, ast.Call):
                            tool_arg = lambda_fn.body.args[-1]
                            if isinstance(tool_arg, ast.Name):
                                tool_var = tool_arg.id
                                print(f"Registered {agent_name} -> {agent_class} using tools: {tool_var}")
                                self.agents.setdefault(agent_class, {
                                    "type": "agent",
                                    "tools": [],
                                    "inherits": []
                                })
                                self.agents[agent_class]["tools"].append(tool_var)
        
        self.generic_visit(node)

    def _get_id(self, node):
        return node.id if isinstance(node, ast.Name) else None

    def _get_str(self, node):
        return node.value if isinstance(node, ast.Constant) and isinstance(node.value, str) else None


def _read_file(file_path: str) -> str:
    """
    Reads the content of a file and returns it as a string.
    """

    try:
        with open(file_path, "r", encoding='utf-8') as file:
            return file.read()
    
    except Exception as e:
        print(f"Error reading file {file_path}: {e}")
        return ""

def build_graph_data(visitor: AutoGenMapper):
    nodes = []
    edges = []

    # Per-visitor: Add Start and End
    nodes.append({
        "name": "Start",
        "function_name": None,
        "docstring": None,
        "node_type": "Start",
        "source_location": None,
        "metadata": {}
    })
    nodes.append({
        "name": "End",
        "function_name": None,
        "docstring": None,
        "node_type": "End",
        "source_location": None,
        "metadata": {}
    })

    for agent_name, agent_data in visitor.agents.items():
        # Agent node
        nodes.append({
            "name": agent_name,
            "function_name": "Assistant",
            "docstring": None,
            "node_type": "Agent",
            "source_location": None,
            "metadata": {}
        })

        # ⬅️ These stay inside build_graph_data to avoid merging
        edges.append({
            "source": "Start",
            "target": agent_name,
            "condition": { "type": "static" },
            "metadata": { "reason": "per-visitor entry point" }
        })
        edges.append({
            "source": agent_name,
            "target": "End",
            "condition": { "type": "static" },
            "metadata": { "reason": "per-visitor exit point" }
        })

        # Tool nodes and edges
        for tool_var in agent_data["tools"]:
            tool_funcs = visitor.tools.get(tool_var, [])
            for func_name in tool_funcs:
                tool_meta = visitor.functions.get(func_name, {})
                tool_node = {
                    "name": f"{agent_name}_{func_name}",
                    "function_name": func_name,
                    "docstring": tool_meta.get("docstring"),
                    "node_type": "Tool",
                    "source_location": tool_meta.get("source_location"),
                    "metadata": {}
                }
                nodes.append(tool_node)

                edges.append({
                    "source": agent_name,
                    "target": tool_node["name"],
                    "condition": { "type": "static" },
                    "metadata": {}
                })

    # Tool-to-tool edges (unchanged)
    all_tools = [f for funcs in visitor.tools.values() for f in funcs]

    for source_func, called_funcs in visitor.tool_calls.items():
        if source_func not in all_tools:
            continue
        for target_func in called_funcs:
            if target_func in all_tools:
                for agent_name, agent_data in visitor.agents.items():
                    for tool_var in agent_data.get("tools", []):
                        if (
                            source_func in visitor.tools.get(tool_var, []) and
                            target_func in visitor.tools.get(tool_var, [])
                        ):
                            edges.append({
                                "source": f"{agent_name}_{source_func}",
                                "target": f"{agent_name}_{target_func}",
                                "condition": { "type": "internal_call" },
                                "metadata": { "reason": f"{source_func} calls {target_func}" }
                            })

    return {
        "nodes": nodes,
        "edges": edges
    }



def merge_visitors(visitors: list[AutoGenMapper]) -> AutoGenMapper:
    merged = AutoGenMapper()

    for visitor in visitors:
        # Merge agents deeply
        for agent_name, agent_data in visitor.agents.items():
            if agent_name not in merged.agents:
                merged.agents[agent_name] = {
                    "type": agent_data["type"],
                    "tools": list(agent_data["tools"]),
                    "inherits": list(agent_data["inherits"]),
                }
            else:
                merged.agents[agent_name]["tools"].extend(
                    t for t in agent_data["tools"] if t not in merged.agents[agent_name]["tools"]
                )
                merged.agents[agent_name]["inherits"].extend(
                    i for i in agent_data["inherits"] if i not in merged.agents[agent_name]["inherits"]
                )

        # Merge tools (tool var → list of function names)
        for tool_var, tool_funcs in visitor.tools.items():
            if tool_var not in merged.tools:
                merged.tools[tool_var] = list(tool_funcs)
            else:
                merged.tools[tool_var].extend(
                    func for func in tool_funcs if func not in merged.tools[tool_var]
                )

        # Merge function definitions (last one wins, safe unless exact name collision)
        merged.functions.update(visitor.functions)

        # Merge tool calls
        for func, called in visitor.tool_calls.items():
            if func not in merged.tool_calls:
                merged.tool_calls[func] = list(called)
            else:
                merged.tool_calls[func].extend(
                    c for c in called if c not in merged.tool_calls[func]
                )

    return merged



def extract_autogen_graph(target_directory: str, output_filename: str):
    """
    Extracts the AutoGen agentic graph structure from a python project and saves it as a JSON file.
    """

    visitors = []

    # Process each Python file
    for root, _, files in os.walk(target_directory):
        for filename in files:

            # Check if the file is a Python file
            if filename.endswith(".py"):
                filepath = os.path.join(root, filename)

                file_content = _read_file(filepath)

                if not file_content:
                    continue
                
                # Parse the file content into an AST & create a visitor
                try:
                    tree = ast.parse(file_content)

                    visitor = AutoGenMapper(filepath)
                    
                    # Visit the AST nodes
                    visitor.visit(tree)
                    visitors.append(visitor)
                
                except SyntaxError as e:
                    print(f"Syntax error in file {filepath}: {e}")
                    continue
    
    merged_visitor = merge_visitors(visitors)
    graph_data = build_graph_data(merged_visitor)

    if graph_data:
        graph_data["metadata"] = {
            "framework": "AutoGen Core",
        }

    # Write to output file
    try:
        with open(output_filename, "w", encoding='utf-8') as outfile:
            json.dump(graph_data, outfile, indent=4)
        print(f"AutoGen graph data written to {output_filename}")
    except Exception as e:
        print(f"Error writing JSON output to {output_filename}: {e}")
        sys.exit(1)
    
    # Success exit
    sys.exit(0)


# --- Main Execution Block ---
if __name__ == "__main__":
    
    parser = argparse.ArgumentParser(description="Extract AutoGen structure from Python files into a LangGraph-compatible format.")
    parser.add_argument("--directory", "-d", type=str, default=".", help="Directory to search for Python files")
    parser.add_argument("--output", "-o", type=str, default="autogen_graph.json", help="Output JSON file path")
    
    args = parser.parse_args()
    
    target_directory = args.directory
    output_filename = args.output

    extract_autogen_graph(target_directory, output_filename)




import ast
import os
import json
import sys


class SwarmMapper(ast.NodeVisitor):
    def __init__(self, filename=None):
        self.current_filename = filename
        self.agents = {}
        self.tools = {}
        self.functions = {}
        self.agent_routing_edges = []

        self.function_returns = {}  # function_name → returned variable
        self.tool_function_to_agent = {}  # func → agent_var (built after parsing)


    def visit_Assign(self, node):
        if isinstance(node.value, ast.Call) and hasattr(node.value.func, 'id'):
            func_id = node.value.func.id

            # Detect Agent(...)
            if func_id == "Agent":
                kwargs = {kw.arg: kw.value for kw in node.value.keywords}
                name = self._get_constant(kwargs.get("name"))
                functions = kwargs.get("functions", [])

                if name:
                    agent_var = node.targets[0].id if isinstance(node.targets[0], ast.Name) else name
                    
                    if agent_var:
                        self.current_agent_var = agent_var
                    self.agents[name] = {
                        "var": agent_var,
                        "type": "Agent",
                        "tools": self._extract_tool_names(functions),
                        "inherits": ["Agent"]
                    }
                

            # Detect Triage Agent
            elif func_id == "create_triage_agent":
                kwargs = {kw.arg: kw.value for kw in node.value.keywords}
                name = self._get_constant(kwargs.get("name"))
                agents_list = kwargs.get("agents")

                triage_var = node.targets[0].id if isinstance(node.targets[0], ast.Name) else name
                self.agents[name] = {
                    "var": triage_var,
                    "type": "TriageAgent",
                    "tools": [],
                    "inherits": ["create_triage_agent"]
                }

                if isinstance(agents_list, ast.List):
                    for agent in agents_list.elts:
                        if isinstance(agent, ast.Name):
                            target_var = agent.id
                            self.agent_routing_edges.append((name, target_var))

        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self._register_function_metadata(node)

        returned_var = None
        for stmt in ast.walk(node):
            if isinstance(stmt, ast.Return) and isinstance(stmt.value, ast.Name):
                returned_var = stmt.value.id

        if returned_var:
            self.function_returns[node.name] = returned_var

        self.generic_visit(node)


    def visit_AsyncFunctionDef(self, node):
        self._register_function_metadata(node)

    def _get_constant(self, node):
        return node.value if isinstance(node, ast.Constant) else None

    def _extract_tool_names(self, tool_list_node):
        tools = []
        if isinstance(tool_list_node, ast.List):
            for elt in tool_list_node.elts:
                if isinstance(elt, ast.Name):
                    tools.append(elt.id)
        return tools

    def _register_function_metadata(self, node):
        self.functions[node.name] = {
            "docstring": ast.get_docstring(node),
            "source_location": {
                "file": self.current_filename,
                "line": node.lineno,
                "col": node.col_offset,
                "end_line": getattr(node, "end_lineno", node.lineno),
                "end_col": getattr(node, "end_col_offset", node.col_offset)
            }
        }


def build_graph_data(visitor):
    nodes = []
    edges = []

    for agent_name, agent_data in visitor.agents.items():
        nodes.append({
            "name": agent_name,
            "function_name": "Assistant",
            "docstring": None,
            "node_type": "Agent",
            "source_location": None,
            "metadata": {"inherits": agent_data["inherits"]}
        })

        edges.append({"source": "Start", "target": agent_name, "condition": {"type": "static"}, "metadata": {}})
        edges.append({"source": agent_name, "target": "End", "condition": {"type": "static"}, "metadata": {}})

        for func_name in agent_data["tools"]:
            tool_meta = visitor.functions.get(func_name, {})
            tool_node_name = f"{agent_name}_{func_name}"

            nodes.append({
                "name": tool_node_name,
                "function_name": func_name,
                "docstring": tool_meta.get("docstring"),
                "node_type": "Tool",
                "source_location": tool_meta.get("source_location"),
                "metadata": {}
            })

            # Only add static Agent → Tool edge if NOT already used in routing
            is_used_in_routing = any(
                len(route) == 3 and route[0] == agent_name and route[2] == func_name
                for route in visitor.agent_routing_edges
            )

            if not is_used_in_routing:
                edges.append({
                    "source": agent_name,
                    "target": tool_node_name,
                    "condition": {"type": "static"},
                    "metadata": {}
                })

    # Build reverse mapping from var name → agent display name
    var_to_agent_name = {v["var"]: k for k, v in visitor.agents.items()}

    for routing in visitor.agent_routing_edges:
        if len(routing) == 3:
            source_agent, target_agent_name, via_func = routing
            intermediate_tool_node = f"{source_agent}_{via_func}"

            edges.append({
                "source": source_agent,
                "target": intermediate_tool_node,
                "condition": {"type": "static"},
                "metadata": {}
            })

            edges.append({
                "source": intermediate_tool_node,
                "target": target_agent_name,
                "condition": {"type": "route"},
                "metadata": {"via": "function_return"}
            })

        elif len(routing) == 2:
            # Legacy route, still allowed (e.g. create_triage_agent)
            source_agent, target_var = routing
            target_agent_name = var_to_agent_name.get(target_var, target_var)

            edges.append({
                "source": source_agent,
                "target": target_agent_name,
                "condition": {"type": "route"},
                "metadata": {"via": "create_triage_agent"}
            })

    return {"nodes": nodes, "edges": edges}



def _read_file(path):
    try:
        with open(path, "r", encoding="utf-8") as f:
            return f.read()
    except Exception as e:
        print(f"Failed to read {path}: {e}")
        return ""


def extract_swarm_graph(directory, output_file):
    visitors = []

    for root, _, files in os.walk(directory):
        for filename in files:
            if filename.endswith(".py"):
                filepath = os.path.join(root, filename)
                content = _read_file(filepath)
                if not content:
                    continue

                try:
                    tree = ast.parse(content)
                    visitor = SwarmMapper(filepath)
                    visitor.visit(tree)
                    visitors.append(visitor)
                except SyntaxError as e:
                    print(f"Syntax error in {filepath}: {e}")
                    continue

    # After visitors are collected
    # Build agent var map: var_name → display_name
    agent_var_to_name = {}
    for visitor in visitors:
        for name, agent in visitor.agents.items():
            agent_var_to_name[agent["var"]] = name
            # Reverse map: tool func → agent_var
            for func in agent["tools"]:
                visitor.tool_function_to_agent[func] = agent["var"]

    # Now resolve function returns to agents
    for visitor in visitors:
        for func, returned_var in visitor.function_returns.items():
            source_agent_var = visitor.tool_function_to_agent.get(func)
            target_agent_name = agent_var_to_name.get(returned_var)
            source_agent_name = agent_var_to_name.get(source_agent_var)

            if source_agent_name and target_agent_name:
                visitor.agent_routing_edges.append(
                    (source_agent_name, target_agent_name, func)
                )


    all_nodes, all_edges = [], []

    # Add a single shared Start and End node
    all_nodes.insert(0, {"name": "Start", "function_name": None, "node_type": "Start", "docstring": None, "source_location": None, "metadata": {}})
    all_nodes.append({"name": "End", "function_name": None, "node_type": "End", "docstring": None, "source_location": None, "metadata": {}})

    for visitor in visitors:
        graph = build_graph_data(visitor)
        all_nodes.extend(graph["nodes"])
        all_edges.extend(graph["edges"])

    graph_data = {
        "metadata": {
            "framework":"OpenAI Swarm"
        },  
        "nodes": all_nodes,
        "edges": all_edges
    }

    try:
        with open(output_file, "w", encoding="utf-8") as f:
            json.dump(graph_data, f, indent=4)
        print(f"[✓] Swarm agent graph written to {output_file}")
    except Exception as e:
        print(f"Error writing JSON output: {e}")
        sys.exit(1)


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--directory", "-d", type=str, default=".", help="Path to Swarm code")
    parser.add_argument("--output", "-o", type=str, default="swarm_graph.json", help="Path to output JSON")
    args = parser.parse_args()

    extract_swarm_graph(args.directory, args.output)



import argparse
import ast
import json
import os
from pathlib import Path
import re
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Union

def get_func_name(node: Union[ast.Name, ast.Attribute]) -> str:
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        base = get_func_name(node.value)
        return f"{base}.{node.attr}" if base else node.attr
    return ""

class FunctionAnalyzer(ast.NodeVisitor):
    def __init__(self, outer_visitor, filename: str, source_node_name: Optional[str] = None, call_site_context: Optional[Dict[str, ast.AST]] = None, depth=0):
        self.outer_visitor = outer_visitor
        self.filename = filename
        self.source_node_name = source_node_name
        self.call_site_context = call_site_context or {}
        self.depth = depth
        self.max_depth = 3

        self.variables: Dict[str, Set[str]] = {}
        self.potential_returns: Set[str] = set()
        self.potential_gotos: Dict[str, Dict[str, Any]] = {}
        self.unresolved_reasons: Set[Tuple[str, str]] = set()
        self.goto_source_locations: Dict[str, Dict[str, Any]] = {}

        self.calls_llm = False
        self.calls_search = False
        self.calls_db = False
        self.calls_tool = False
        self.is_router = False

    def _get_location_dict(self, node: ast.AST) -> Dict[str, Any]:
        return {
            "file": self.filename,
            "line": node.lineno,
            "col": node.col_offset,
            "end_line": getattr(node, 'end_lineno', None),
            "end_col": getattr(node, 'end_col_offset', None),
        }

    def _build_call_context(self, func_def: Union[ast.FunctionDef, ast.AsyncFunctionDef], call_node: ast.Call) -> Dict[str, ast.AST]:
        context = {}
        params = func_def.args.args
        param_names = [p.arg for p in params]
        for i, arg_node in enumerate(call_node.args):
            if i < len(param_names): context[param_names[i]] = arg_node
        for kw in call_node.keywords:
            if kw.arg in param_names: context[kw.arg] = kw.value
        return context

    def _analyze_called_function(self, func_call_node: ast.Call) -> Tuple[Set[str], str, Optional[str]]:
        if self.depth >= self.max_depth:
            func_name_str = self.outer_visitor._stringify_ast_node(func_call_node.func)
            return set(), "unresolved_call_max_depth", func_name_str

        func_simple_name = get_func_name(func_call_node.func).split('.')[-1]
        func_def = self.outer_visitor.function_defs.get(func_simple_name)

        if func_def:
            inner_context = self._build_call_context(func_def, func_call_node)
            inner_analyzer = FunctionAnalyzer(self.outer_visitor, self.filename, call_site_context=inner_context, depth=self.depth + 1)
            try:
                inner_analyzer.visit(func_def)
                resolved_returns = {ret for ret in inner_analyzer.potential_returns if not ret.startswith("variable(") and not ret.startswith("unresolved_")}
                has_unresolved = bool(inner_analyzer.potential_returns - resolved_returns) or bool(inner_analyzer.unresolved_reasons)
                status = "resolved_partial" if resolved_returns and has_unresolved else ("resolved" if resolved_returns else "unresolved_no_returns")
                unresolved_detail = f"inner_call:{func_simple_name}" if has_unresolved else None
                return resolved_returns, status, unresolved_detail
            except Exception as e:
                return set(), "unresolved_call_error", func_simple_name
        else:
            return set(), "unresolved_call_not_found", func_simple_name

    def visit_Assign(self, node: ast.Assign):
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            target_var = node.targets[0].id
            value_node = node.value
            new_values: Set[str] = set()
            # resolution_status = "unknown"
            # resolution_detail = None

            if isinstance(value_node, ast.Constant) and isinstance(value_node.value, str):
                new_values.add(value_node.value)
                # resolution_status = "resolved_literal"
            elif isinstance(value_node, ast.Name):
                source_var = value_node.id
                if source_var in self.variables:
                    new_values.update(self.variables[source_var])
                    resolution_status = "resolved_variable"
                    resolution_detail = source_var
                elif source_var in self.call_site_context:
                    arg_node = self.call_site_context[source_var]
                    resolved_arg = self.outer_visitor.extract_argument_value(arg_node)
                    if isinstance(resolved_arg, str):
                         new_values.add(resolved_arg)
                         resolution_status = "resolved_parameter"
                         resolution_detail = source_var
                    else:
                         resolution_status = "unresolved_parameter_type"
                         resolution_detail = source_var
                         self.unresolved_reasons.add(('parameter', source_var))
                         new_values.add(f"unresolved_param({source_var})")
                else:
                     resolution_status = "unresolved_variable"
                     resolution_detail = source_var
                     self.unresolved_reasons.add(('variable', source_var))
                     new_values.add(f"variable({source_var})")

            elif isinstance(value_node, ast.Call):
                 call_returns, status, detail = self._analyze_called_function(value_node)
                 new_values.update(call_returns)
                 resolution_status = status
                 resolution_detail = detail
                 if status.startswith("unresolved"):
                      self.unresolved_reasons.add(('call_result', detail or self.outer_visitor._stringify_ast_node(value_node.func)))


            if target_var not in self.variables:
                self.variables[target_var] = set()
            if new_values:
                 self.variables[target_var].update(new_values)
            # TODO: Track resolution status per variable? Complex.

        self.generic_visit(node)

    def visit_Return(self, node: ast.Return):
        if node.value:
            value_node = node.value
            current_potential_returns: Set[str] = set()
            resolution_status = "unknown"
            resolution_detail = None
            goto_target_results : Dict[str, Dict[str, Any]] = {}

            # --- Resolve the value being returned (for direct returns) ---
            if isinstance(value_node, ast.Constant) and isinstance(value_node.value, str):
                current_potential_returns.add(value_node.value)
                resolution_status = "resolved_literal"
            elif isinstance(value_node, ast.Name):
                var_name = value_node.id
                if var_name in self.call_site_context:
                    arg_node = self.call_site_context[var_name]
                    resolved_arg = self.outer_visitor.extract_argument_value(arg_node)
                    if isinstance(resolved_arg, str):
                        current_potential_returns.add(resolved_arg)
                        resolution_status = "resolved_parameter"
                        resolution_detail = var_name
                    else:
                        resolution_status = "unresolved_parameter_type"
                        resolution_detail = var_name
                        self.unresolved_reasons.add(('parameter', var_name))
                        current_potential_returns.add(f"unresolved_param({var_name})")
                elif var_name in self.variables:
                    current_potential_returns.update(self.variables[var_name])
                    resolution_status = "resolved_variable"
                    resolution_detail = var_name
                    if any(v.startswith("unresolved") or v.startswith("variable") for v in self.variables[var_name]):
                        resolution_status = "resolved_variable_partial"
                        self.unresolved_reasons.add(('variable_propagation', var_name))
                else:
                    resolution_status = "unresolved_variable"
                    resolution_detail = var_name
                    self.unresolved_reasons.add(('variable', var_name))
                    current_potential_returns.add(f"variable({var_name})")

            self.potential_returns.update(r for r in current_potential_returns if not r.startswith('unresolved') and not r.startswith('variable'))
            if len(current_potential_returns) > 1:
                self.is_router = True


            # --- Check for Command(goto=...) ---
            if isinstance(value_node, ast.Call):
                call = value_node
                resolved_func_fqn = self.outer_visitor._resolve_fq_name(call.func)
                is_command_call = resolved_func_fqn and resolved_func_fqn.endswith('Command')

                if resolved_func_fqn:
                    if re.search(r'\b(llm|chat|openai|anthropic|gemini|generate)\b', resolved_func_fqn, re.IGNORECASE):
                        self.calls_llm = True
                    if re.search(r'\b(search|retrieve|tavily|google|bing)\b', resolved_func_fqn, re.IGNORECASE):
                        self.calls_search = True
                    if re.search(r'\b(sql|database|db|query)\b', resolved_func_fqn, re.IGNORECASE):
                        self.calls_db = True
                    if re.search(r'\b(tool|executor)\b', resolved_func_fqn, re.IGNORECASE):
                         self.calls_tool = True

                if is_command_call:
                    for keyword in call.keywords:
                        if keyword.arg == 'goto':
                            goto_value_node = keyword.value
                            goto_target_results = self._resolve_goto_value(goto_value_node)

                            for target, res_info in goto_target_results.items():
                                self.potential_gotos[target] = res_info
                                self.goto_source_locations[target] = self._get_location_dict(node)
                                if res_info.get("status", "").startswith("unresolved"):
                                    self.unresolved_reasons.add((res_info.get("type", "goto"), res_info.get("detail", "?")))
                            self.is_router = True
                            break

        self.generic_visit(node)

    def _resolve_goto_value(self, value_node: ast.AST) -> Dict[str, Dict[str, Any]]:
        targets: Dict[str, Dict[str, Any]] = {}


        def add_target(target_str: str, status: str, type_: str, detail: Optional[str]):
            actual_target = "End" if target_str == "END" else target_str
            if actual_target:
                targets[actual_target] = {"status": status, "type": type_, "detail": detail}
                if actual_target == "End":
                     self.outer_visitor._add_node_if_not_exists("End", "Implicit END node", {}, None)

        if isinstance(value_node, ast.Constant) and isinstance(value_node.value, str):
            add_target(value_node.value, "resolved_literal", "literal", None)
        elif isinstance(value_node, ast.Name):
            var_name = value_node.id
            if var_name == "END":
                add_target(var_name, "resolved_literal", "literal", "END Constant")
            elif var_name in self.variables:
                possible_values = self.variables[var_name]
                has_unresolved_source = False
                resolved_count = 0
                for val in possible_values:
                    if val.startswith(("variable(", "unresolved_")):
                        has_unresolved_source = True
                        self.unresolved_reasons.add(('variable_propagation', var_name))
                    else:
                        status = "resolved_variable_partial" if has_unresolved_source else "resolved_variable"
                        add_target(val, status, "variable", var_name)
                        resolved_count += 1
                if resolved_count == 0 and has_unresolved_source:
                     targets["?"] = {"status": "unresolved_variable_source", "type": "variable", "detail": var_name}

            elif var_name in self.call_site_context:
                 arg_node = self.call_site_context[var_name]
                 resolved_arg = self.outer_visitor.extract_argument_value(arg_node)
                 if isinstance(resolved_arg, str):
                     add_target(resolved_arg, "resolved_parameter", "parameter", var_name)
                 else:
                      add_target("?", "unresolved_parameter_type", "parameter", var_name)
                      self.unresolved_reasons.add(('parameter', var_name))
            else:
                add_target("?", "unresolved_variable", "variable", var_name)
                self.unresolved_reasons.add(('variable', var_name))

        elif isinstance(value_node, ast.List):
            for i, element in enumerate(value_node.elts):
                element_targets = self._resolve_goto_value(element)
                for target, res_info in element_targets.items():
                    res_info["type"] = "list_element"
                    res_info["detail"] = f"{res_info.get('detail', '')}[{i}]" if res_info.get('detail') else f"index {i}"
                    targets[target] = res_info

        elif isinstance(value_node, ast.Call):
             call_returns, status, detail = self._analyze_called_function(value_node)
             if call_returns:
                 for ret_val in call_returns:
                     add_target(ret_val, status, "call_result", detail or self.outer_visitor._stringify_ast_node(value_node.func))
             elif status.startswith("unresolved"):
                  add_target("?", status, "call_result", detail or self.outer_visitor._stringify_ast_node(value_node.func))
                  self.unresolved_reasons.add(('call_result', detail or self.outer_visitor._stringify_ast_node(value_node.func)))
             else:
                 add_target("?", "unresolved_call_no_string_return", "call_result", detail or self.outer_visitor._stringify_ast_node(value_node.func))
                 self.unresolved_reasons.add(('call_result_no_string', detail or self.outer_visitor._stringify_ast_node(value_node.func)))
        else:
             unrec_type = type(value_node).__name__
             add_target("?", f"unresolved_goto_value_type", unrec_type, self.outer_visitor._stringify_ast_node(value_node))
             self.unresolved_reasons.add(('goto_value_type', unrec_type))

        return targets


# --- Enhanced GraphVisitor ---

class GraphVisitor(ast.NodeVisitor):
    def __init__(self):
        self.nodes: List[Dict[str, Any]] = []
        self.edges: List[Dict[str, Any]] = []
        self.node_names: Set[str] = set()
        self.function_defs: Dict[str, Union[ast.FunctionDef, ast.AsyncFunctionDef]] = {}
        self.node_function_map: Dict[str, str] = {}
        self.graph_name: Optional[str] = None
        self.current_filename: str = ""

        self.graph_class_fqcn = "langgraph.graph.StateGraph"
        self.command_class_fqn = "langgraph.types.Command"

        self.import_aliases: Dict[str, str] = {}
        self.import_aliases_fully: Dict[str, str] = {}
        self.variable_is_target_instance: Dict[str, bool] = {}
        self.variable_values: Dict[str, Union[ast.List, ast.Dict, ast.Constant]] = {}

        self.has_start = False
        self.has_end = False

    def _get_location_dict(self, node: ast.AST) -> Dict[str, Any]:
        if not self.current_filename or not hasattr(node, 'lineno'):
             return None
        return {
            "file": self.current_filename,
            "line": node.lineno,
            "col": node.col_offset,
            "end_line": getattr(node, 'end_lineno', None),
            "end_col": getattr(node, 'end_col_offset', None),
        }

    def visit_file(self, tree: ast.AST, filename: str):
        self.current_filename = filename
        self.visit(tree)
        self.current_filename = ""

    def visit_Import(self, node: ast.Import):
        for alias in node.names:
            local_name = alias.asname if alias.asname else alias.name
            self.import_aliases[local_name] = alias.name
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module is None:
            self.generic_visit(node)
            return
        base_module = node.module
        for alias in node.names:
            local_name = alias.asname if alias.asname else alias.name
            imported_name = alias.name
            self.import_aliases_fully[local_name] = f"{base_module}.{imported_name}"
            if '.' in base_module:
                 prefix = base_module.split('.')[0]
                 if prefix not in self.import_aliases: self.import_aliases[prefix] = prefix
            elif base_module not in self.import_aliases: self.import_aliases[base_module] = base_module
        self.generic_visit(node)

    def visit_FunctionDef(self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):
        self.function_defs[node.name] = node
        self.generic_visit(node)

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_Assign(self, node: ast.Assign):
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            var_name = node.targets[0].id
            value = node.value
            if isinstance(value, ast.Call):
                resolved_fqn = self._resolve_fq_name(value.func)
                if resolved_fqn == self.graph_class_fqcn:
                    self.variable_is_target_instance[var_name] = True
                    if self.graph_name is None:
                        self.graph_name = var_name
            if isinstance(value, (ast.Constant, ast.List, ast.Dict)):
                self.variable_values[var_name] = value
            else:
                if var_name in self.variable_values: del self.variable_values[var_name]
            if isinstance(value, ast.Name) and value.id in self.variable_is_target_instance:
                 self.variable_is_target_instance[var_name] = True
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name):
            instance_name = node.func.value.id
            if self.variable_is_target_instance.get(instance_name, False):
                method_name = node.func.attr
                processor = getattr(self, f"process_{method_name}", None)
                if processor:
                    try: processor(node)
                    except Exception as e: print(f"Error processing {method_name} call at {self._get_location_dict(node)}: {e}")
        self.generic_visit(node)

    def _add_node_if_not_exists(self, name, function_name, metadata, location_dict):
        node_name_to_add = "Start" if name == "START" else "End" if name == "END" else name
        is_implicit = name in ["START", "END" , "__start__" , "__end__"]
        if node_name_to_add not in self.node_names:
            node_data = {
                "name": node_name_to_add,
                "function_name": function_name if not is_implicit else None,
                "docstring": None,
                "node_type": "Special" if is_implicit else "Unknown",
                "source_location": location_dict if not is_implicit else None,
                "metadata": metadata or {}
            }

            if not is_implicit and function_name:
                 func_def = self.function_defs.get(function_name)
                 print(func_def)
                 if func_def:
                     try:
                          docstring = ast.get_docstring(func_def)
                          print(docstring)
                          node_data["docstring"] = docstring
                          node_data["node_type"] = self._classify_node(node_name_to_add, function_name, func_def)
                     except Exception as e:
                          print(f"Warning: Could not process function '{function_name}' for node '{node_name_to_add}': {e}")
                 else :
                     try :
                          node_data["node_type"] = self._classify_node(node_name_to_add, function_name, None)
                     except Exception as e:
                          print(f"Warning: Could not process function '{function_name}' for node '{node_name_to_add}': {e}")


            self.nodes.append(node_data)
            self.node_names.add(node_name_to_add)
            if node_name_to_add == "Start": self.has_start = True
            if node_name_to_add == "End": self.has_end = True


    def _classify_node(self, node_name: str, function_name: Optional[str], func_def: Optional[Union[ast.FunctionDef, ast.AsyncFunctionDef]]) -> str:
        if not function_name and not node_name: return "Unknown"

        check_name = node_name.lower() if node_name else ""
        check_func = function_name.lower() if function_name else ""
        combined_check = f"{check_name} {check_func}".strip()

        if check_name in ["start", "start_node", "__start__", "begin", "begin_node"]:
            return "Start"
        if check_name in ["input", "input_node", "entry", "entry_point"]:
            return "Input"
        if check_name in ["output", "output_node", "__end__", "end", "finish", "exit", "finish_point"]:
            return "Output"
        if check_name in ["tool_node", "tools", "tool_executor", "actions", "action_node"]:
            return "ToolExecutor"

        if re.search(r'\b(router|route|routing|conditional|condition|branch|switch|decide|decision|choose|selector|map_step|path_map)\b', combined_check):
            return "Router"

        if re.search(r'\b(agent|llm|openai|anthropic|gemini|llama|mistral|claude|chat|generate|invoke|call_model|prompt|predict|reasoning|step_back)\b', combined_check):
            return "Agent"

        if re.search(r'\b(plan|planner|planning|sequence|orchestrate|orchestrator)\b', combined_check):
            return "Planner"

        if re.search(r'\b(search|retrieve|retriever|tavily|google|bing|duckduckgo|arxiv|wikipedia|lookup|knowledge)\b', combined_check):
            return "Search/Retriever"

        is_tool_related = (
            re.search(r'\b(tool(?!_node)|action(?!_node)|api|function_call|calculator|python_repl|terminal)\b', combined_check) or
            "tool" in check_name.split("_") or
            "tools" in check_name.split("_") or
            re.search(r'[_-]tools?\b', check_name)
        )

        if is_tool_related and not re.search(r'\b(search|retrieve|database|sql)\b', combined_check):
             if func_def:
                analyzer = FunctionAnalyzer(self, getattr(func_def, '_filename', self.current_filename), source_node_name=node_name)
                try:
                    analyzer.visit(func_def)
                    if analyzer.calls_tool: return "Tool"
                except Exception: pass
             return "Tool"


        if re.search(r'\b(db|database|sql|query|vectorstore|storage|memory|persist)\b', combined_check):
             if re.search(r'\b(memory|context|history)\b', combined_check) and not re.search(r'\b(sql|query|vectorstore)\b', combined_check):
                 return "MemoryManagement"
             else:
                 return "Database/Storage"

        if re.search(r'\b(reflect|reflection|review|critique|critic|correct|refine|revise|validate|check)\b', combined_check):
            return "Reflection/Correction"

        if re.search(r'\b(input|entry|receive|user_input|question)\b', combined_check):
            return "Input"
        if re.search(r'\b(output|exit|finish|display|respond|response|answer|result)\b', combined_check):
            return "Output"


        if func_def:
            func_filename = getattr(func_def, '_filename', self.current_filename)
            analyzer = FunctionAnalyzer(self, func_filename, source_node_name=node_name)
            try:
                analyzer.visit(func_def)
                if analyzer.is_router: return "Router"
                if analyzer.calls_llm: return "Agent/LLM"
                if analyzer.calls_search: return "Search/Retriever"
                if analyzer.calls_db: return "Database/Storage"
                if analyzer.calls_tool: return "Tool"
            except Exception as e:
                 pass

        print(node_name)
        return "Generic"

    def _resolve_fq_name(self, node: ast.AST) -> Optional[str]:
        if isinstance(node, ast.Name):
            node_id = node.id
            if node_id in self.import_aliases_fully: return self.import_aliases_fully[node_id]
            return node_id
        elif isinstance(node, ast.Attribute):
            base_node = node.value
            attr_name = node.attr
            base_fqn = self._resolve_fq_name(base_node)
            if base_fqn:
                if base_fqn in self.import_aliases: return f"{self.import_aliases[base_fqn]}.{attr_name}"
                else: return f"{base_fqn}.{attr_name}"
            else: return None
        elif isinstance(node, ast.Constant) and isinstance(node.value, str): return node.value
        return None

    def _stringify_ast_node(self, node: ast.AST) -> str:
        if isinstance(node, ast.Constant): return repr(node.value)
        elif isinstance(node, ast.Name): return node.id
        else:
            try: return ast.dump(node)
            except Exception: return f"<{type(node).__name__}>"

    # --- Process Methods (Enhanced) ---

    def process_add_node(self, node: ast.Call):
        node_name = None
        action_ref = None
        metadata = {}
        pos_args = node.args
        kw_args = {kw.arg: kw.value for kw in node.keywords if kw.arg}
        location = self._get_location_dict(node)


        if 'node' in kw_args: node_name = self.extract_argument_value(kw_args['node'])
        if 'action' in kw_args: action_ref = self.extract_argument_value(kw_args['action'])
        if len(pos_args) >= 1 and node_name is None:
             arg0_val = self.extract_argument_value(pos_args[0])
             if isinstance(pos_args[0], ast.Constant) and isinstance(arg0_val, str):
                 node_name = arg0_val
                 if len(pos_args) >= 2 and action_ref is None: action_ref = self.extract_argument_value(pos_args[1])
             elif isinstance(pos_args[0], ast.Name) or isinstance(pos_args[0], ast.Attribute):
                 action_ref = self.extract_argument_value(pos_args[0])
                 node_name = action_ref

        if 'metadata' in kw_args and isinstance(kw_args['metadata'], ast.Dict):
            try: metadata = {self.extract_argument_value(k): self.extract_argument_value(v) for k, v in zip(kw_args['metadata'].keys, kw_args['metadata'].values)}
            except Exception: metadata = {"error": "Cannot parse metadata dict"}

        if node_name and isinstance(node_name, str):
            actual_func_name = action_ref if isinstance(action_ref, str) else node_name
            print(action_ref)
            self._add_node_if_not_exists(node_name, actual_func_name, metadata, location)
            if actual_func_name:
                self.node_function_map[node_name] = actual_func_name
        else:
            print(f"Warning: Could not determine node name for add_node call at {location}")

    def process_add_edge(self, node: ast.Call):
        source = None
        target = None
        pos_args = node.args
        kw_args = {kw.arg: kw.value for kw in node.keywords if kw.arg}
        location = self._get_location_dict(node)

        if 'start_key' in kw_args: source = self.extract_argument_value(kw_args['start_key'])
        if 'end_key' in kw_args: target = self.extract_argument_value(kw_args['end_key'])
        if source is None and len(pos_args) >= 1: source = self.extract_argument_value(pos_args[0])
        if target is None and len(pos_args) >= 2: target = self.extract_argument_value(pos_args[1])

        if source == "Start": self._add_node_if_not_exists("START", None, {}, None)
        if target == "End": self._add_node_if_not_exists("END", None, {}, None)

        if source and target:
             sources = source if isinstance(source, list) else [source]
             for s_item in sources:
                 s_item_norm = "Start" if s_item == "START" else s_item
                 if s_item_norm == "Start": self._add_node_if_not_exists("START", None, {}, None)
                 if s_item_norm and target:
                     self.edges.append({
                         "source": s_item_norm,
                         "target": target,
                         "condition": {"type": "static"},
                         "metadata": {"definition_location": location}
                     })
        else:
            print(f"Warning: Could not extract source/target from add_edge at {location}")


    def process_add_conditional_edges(self, node: ast.Call):
        source = None
        path_arg = None
        path_map_arg = None
        pos_args = node.args
        kw_args = {kw.arg: kw.value for kw in node.keywords if kw.arg}
        location = self._get_location_dict(node)

        if 'source' in kw_args: source = self.extract_argument_value(kw_args['source'])
        elif len(pos_args) >= 1: source = self.extract_argument_value(pos_args[0])

        if 'path' in kw_args: path_arg = kw_args['path']
        elif len(pos_args) >= 2: path_arg = pos_args[1]

        if 'path_map' in kw_args: path_map_arg = kw_args['path_map']
        elif len(pos_args) >= 3: path_map_arg = pos_args[2]

        if not source:
            print(f"Warning: Missing source in add_conditional_edges at {location}")
            return

        path_func_name = None
        if isinstance(path_arg, ast.Name): path_func_name = path_arg.id
        elif isinstance(path_arg, ast.Attribute): path_func_name = get_func_name(path_arg)

        if path_map_arg:
            resolved_path_map = None
            map_source_detail = ""
            resolution_status = "unknown"

            if isinstance(path_map_arg, ast.Name):
                var_name = path_map_arg.id
                map_source_detail = f"variable: {var_name}"
                if var_name in self.variable_values and isinstance(self.variable_values[var_name], (ast.Dict, ast.List)):
                    resolved_path_map = self.variable_values[var_name]
                    resolution_status = "resolved_variable"
                else:
                    print(f"Warning: Cannot resolve path_map variable '{var_name}' for source '{source}' at {location}.")
                    resolution_status = "unresolved_variable"
                    self.edges.append({
                        "source": source, "target": "?",
                        "condition": {
                            "type": "conditional_map",
                            "value": None,
                            "condition_function": path_func_name,
                            # "resolution": {"status": resolution_status, "detail": map_source_detail}
                        },
                        "metadata": {"definition_location": location}
                    })
                    return
            elif isinstance(path_map_arg, (ast.Dict, ast.List)):
                resolved_path_map = path_map_arg
                resolution_status = "resolved_literal"
                map_source_detail = "literal"
            else:
                 print(f"Warning: Unsupported path_map type '{type(path_map_arg)}' for source '{source}' at {location}.")
                 resolution_status = "unresolved_type"
                 map_source_detail = type(path_map_arg).__name__
                 self.edges.append({ "source": source, "target": "?", "condition": { "type": "conditional_map", "value": None, "condition_function": path_func_name, "resolution": {"status": resolution_status, "detail": map_source_detail}}, "metadata": {"definition_location": location}})
                 return

            if isinstance(resolved_path_map, ast.Dict):
                for key_node, value_node in zip(resolved_path_map.keys, resolved_path_map.values):
                    condition_val = self.extract_argument_value(key_node)
                    target_val = self.extract_argument_value(value_node)
                    if target_val == "End": self._add_node_if_not_exists("END", None, {}, None)
                    if condition_val is not None and target_val:
                        self.edges.append({
                            "source": source, "target": target_val,
                            "condition": {
                                "type": "conditional_map",
                                "value": condition_val,
                                "condition_function": path_func_name,
                                # "resolution": {"status": resolution_status, "detail": map_source_detail}
                            },
                            "metadata": {"definition_location": location}
                        })
            elif isinstance(resolved_path_map, ast.List):
                 for element_node in resolved_path_map.elts:
                     target_val = self.extract_argument_value(element_node)
                     if target_val == "End": self._add_node_if_not_exists("END", None, {}, None)
                     if target_val:
                          self.edges.append({
                              "source": source, "target": target_val,
                              "condition": {
                                  "type": "conditional_map_list",
                                  "value": None,
                                  "condition_function": path_func_name,
                                #   "resolution": {"status": resolution_status, "detail": map_source_detail}
                              },
                              "metadata": {"definition_location": location}
                          })

        elif path_func_name:
             func_def = self.function_defs.get(path_func_name)
             if func_def:
                 analyzer = FunctionAnalyzer(self, self.current_filename)
                 analyzer.visit(func_def)
                 possible_targets = analyzer.potential_returns

                 if not possible_targets and analyzer.unresolved_reasons:
                      detail = ", ".join([f"{t}:{d}" for t, d in analyzer.unresolved_reasons])
                      print("Warning : no target")
                      self.edges.append({
                            "source": source, "target": "?",
                            "condition": {
                                "type": "conditional_func_return",
                                "value": None,
                                "condition_function": path_func_name,
                                # "resolution": {"status": "unresolved_function_analysis", "detail": detail}
                            },
                            "metadata": {"definition_location": location}
                      })
                 elif not possible_targets:
                     print(f"Warning: Path function '{path_func_name}' seems to have no direct string return values at {location}.")
                     self.edges.append({
                           "source": source, "target": "?",
                           "condition": { "type": "conditional_func_return", "value": None, "condition_function": path_func_name, "resolution": {"status": "unresolved_no_returns"}},
                           "metadata": {"definition_location": location}
                     })

                 for target in possible_targets:
                     if target == "End": self._add_node_if_not_exists("END", None, {}, None)
                     if target:
                        self.edges.append({
                            "source": source, "target": target,
                            "condition": {
                                "type": "conditional_func_return",
                                "value": target,
                                "condition_function": path_func_name,
                                # "resolution": {"status": "resolved"}
                            },
                            "metadata": {"definition_location": location, "potential_return": True}
                        })
             else:
                  print(f"Warning: Path function '{path_func_name}' definition not found at {location}.")
                  self.edges.append(
                      {"source": source,
                       "target": "?",
                       "condition": {"type": "conditional_func_return", "value": None,
                                      "condition_function": path_func_name,
                                    # "resolution": {"status": "unresolved_function_not_found"}
                                    },
                                    "metadata": {"definition_location": location}})
        else:
             print(f"Warning: Insufficient arguments for add_conditional_edges from '{source}' at {location}.")


    def process_add_sequence(self, node: ast.Call):
        location = self._get_location_dict(node)
        if len(node.args) == 1 and isinstance(node.args[0], ast.List):
            seq_nodes_ast = node.args[0].elts
            last_node_name = None
            for i, item_ast in enumerate(seq_nodes_ast):
                current_node_name = self.extract_argument_value(item_ast)
                if isinstance(current_node_name, list) and len(current_node_name) > 0 and isinstance(current_node_name[0], str):
                    current_node_name = current_node_name[0]

                if isinstance(current_node_name, str):
                    self._add_node_if_not_exists(current_node_name, current_node_name, {}, self._get_location_dict(item_ast))
                    if last_node_name:
                        self.edges.append({
                            "source": last_node_name,
                            "target": current_node_name,
                            "condition": {"type": "sequence"},
                            "metadata": {"definition_location": location}
                        })
                    last_node_name = current_node_name
                else:
                     print(f"Warning: Could not determine node name for item {i} in add_sequence at {location}.")
                     last_node_name = None
        else: print(f"Warning: add_sequence call malformed at {location}.")


    def process_set_entry_point(self, node: ast.Call):
        target = None
        location = self._get_location_dict(node)
        if len(node.args) == 1: target = self.extract_argument_value(node.args[0])
        for kw in node.keywords:
            if kw.arg == 'key': target = self.extract_argument_value(kw.value)

        if target:
            self._add_node_if_not_exists("START", None, {}, None)
            self.edges.append({
                "source": "Start",
                "target": target,
                "condition": {"type": "entry_point"},
                "metadata": {"definition_location": location}
            })
        else: print(f"Warning: Could not extract target from set_entry_point at {location}.")

    def process_set_conditional_entry_point(self, node: ast.Call):
         path_arg = None
         path_map_arg = None
         pos_args = node.args
         kw_args = {kw.arg: kw.value for kw in node.keywords if kw.arg}
         location = self._get_location_dict(node)
         source = "Start"

         self._add_node_if_not_exists("START", None, {}, None)

         if 'path' in kw_args: path_arg = kw_args['path']
         elif len(pos_args) >= 1: path_arg = pos_args[0]

         if 'path_map' in kw_args: path_map_arg = kw_args['path_map']
         elif len(pos_args) >= 2: path_map_arg = pos_args[1]

         path_func_name = None
         if isinstance(path_arg, ast.Name): path_func_name = path_arg.id
         elif isinstance(path_arg, ast.Attribute): path_func_name = get_func_name(path_arg)

         if path_map_arg:
            resolved_path_map = None
            map_source_detail = ""
            # resolution_status = "unknown"
            if isinstance(path_map_arg, ast.Name):
                 var_name = path_map_arg.id
                 map_source_detail = f"variable: {var_name}"
                 if var_name in self.variable_values and isinstance(self.variable_values[var_name], (ast.Dict, ast.List)):
                      resolved_path_map = self.variable_values[var_name]; resolution_status = "resolved_variable"
                #  else: resolution_status = "unresolved_variable"
            elif isinstance(path_map_arg, (ast.Dict, ast.List)):
                 resolved_path_map = path_map_arg; resolution_status = "resolved_literal"; map_source_detail = "literal"
            else:
                # resolution_status = "unresolved_type";
                map_source_detail = type(path_map_arg).__name__

            # if resolution_status.startswith("unresolved"):
            #      self.edges.append({"source": source, "target": "?", "condition": {"type": "conditional_entry_map", "value": None, "condition_function": path_func_name, "resolution": {"status": resolution_status, "detail": map_source_detail}}, "metadata": {"definition_location": location}})
            #      return

            if isinstance(resolved_path_map, ast.Dict):
                 for key_node, value_node in zip(resolved_path_map.keys, resolved_path_map.values):
                      condition_val = self.extract_argument_value(key_node)
                      target_val = self.extract_argument_value(value_node)
                      if target_val == "End": self._add_node_if_not_exists("END", None, {}, None)
                      if condition_val is not None and target_val: self.edges.append({"source": source, "target": target_val, "condition": {"type": "conditional_entry_map", "value": condition_val, "condition_function": path_func_name, "resolution": {"status": resolution_status, "detail": map_source_detail}}, "metadata": {"definition_location": location}})
            elif isinstance(resolved_path_map, ast.List):
                  for element_node in resolved_path_map.elts:
                       target_val = self.extract_argument_value(element_node)
                       if target_val == "End": self._add_node_if_not_exists("END", None, {}, None)
                       if target_val: self.edges.append({"source": source, "target": target_val, "condition": {"type": "conditional_entry_map_list", "value": None, "condition_function": path_func_name, "resolution": {"status": resolution_status, "detail": map_source_detail}}, "metadata": {"definition_location": location}})

         elif path_func_name:
            func_def = self.function_defs.get(path_func_name)
            if func_def:
                 analyzer = FunctionAnalyzer(self, self.current_filename); analyzer.visit(func_def)
                 possible_targets = analyzer.potential_returns
                 if not possible_targets and analyzer.unresolved_reasons:
                      print("Warning :  no target")
                      detail = ", ".join([f"{t}:{d}" for t, d in analyzer.unresolved_reasons])
                      self.edges.append({"source": source, "target": "?", "condition": {"type": "conditional_entry_func", "value": None, "condition_function": path_func_name, "resolution": {"status": "unresolved_function_analysis", "detail": detail}}, "metadata": {"definition_location": location}})
                 elif not possible_targets: self.edges.append({"source": source, "target": "?", "condition": {"type": "conditional_entry_func", "value": None, "condition_function": path_func_name, "resolution": {"status": "unresolved_no_returns"}}, "metadata": {"definition_location": location}})
                 for target in possible_targets:
                      if target == "End": self._add_node_if_not_exists("END", None, {}, None)
                      if target: self.edges.append({"source": source, "target": target, "condition": {"type": "conditional_entry_func", "value": target, "condition_function": path_func_name, "resolution": {"status": "resolved"}}, "metadata": {"definition_location": location, "potential_return": True}})
            else: self.edges.append({"source": source, "target": "?", "condition": {"type": "conditional_entry_func", "value": None, "condition_function": path_func_name, "resolution": {"status": "unresolved_function_not_found"}}, "metadata": {"definition_location": location}})
         else: print(f"Warning: Insufficient arguments for set_conditional_entry_point at {location}.")


    def process_set_finish_point(self, node: ast.Call):
        source = None
        location = self._get_location_dict(node)
        if len(node.args) == 1: source = self.extract_argument_value(node.args[0])
        for kw in node.keywords:
            if kw.arg == 'key': source = self.extract_argument_value(kw.value)

        if source:
            self._add_node_if_not_exists("END", None, {}, None)
            self.edges.append({
                "source": source,
                "target": "End",
                "condition": {"type": "finish_point"},
                "metadata": {"definition_location": location}
            })
        else: print(f"Warning: Could not extract source from set_finish_point at {location}.")


    def extract_argument_value(self, arg: ast.AST) -> Any:
        if isinstance(arg, ast.Constant):
            value = arg.value
            if value == "START": return "Start"
            if value == "END": return "End"
            return value
        elif isinstance(arg, ast.Name):
            if arg.id == "START": return "Start"
            if arg.id == "END": return "End"
            if arg.id in self.variable_values:
                 return self.extract_argument_value(self.variable_values[arg.id])
            return arg.id
        elif isinstance(arg, ast.List):
            return [self.extract_argument_value(elt) for elt in arg.elts]
        elif isinstance(arg, ast.Dict):
             try:
                 return {self.extract_argument_value(k): self.extract_argument_value(v) for k, v in zip(arg.keys, arg.values)}
             except Exception:
                 return f"<Dict: {self._stringify_ast_node(arg)}>"
        elif isinstance(arg, ast.Attribute):
             fqn = get_func_name(arg)
             if fqn.endswith(".START"): return "Start"
             if fqn.endswith(".END"): return "End"
             return fqn
        elif isinstance(arg, ast.Call):
            fqn = get_func_name(arg.func)
            return fqn
        print( f"<Unsupported: {self._stringify_ast_node(arg)}>")
        return "_"


    def analyze_functions_for_goto(self):
        print("Analyzing functions for dynamic 'goto' edges...")
        analyzed_signatures = set()

        for node_name, function_name in self.node_function_map.items():
            func_def = self.function_defs.get(function_name)
            if func_def:
                func_filename = self.current_filename
                for fn, tree_node in self.function_defs.items():
                    if fn == function_name and hasattr(tree_node, '_filename'):
                        func_filename = tree_node._filename
                        break

                analyzer = FunctionAnalyzer(self, func_filename, source_node_name=node_name)
                try: analyzer.visit(func_def)
                except Exception as e: print(f"  Error analyzing function '{function_name}' for gotos: {e}"); continue

                for target, resolution_info in analyzer.potential_gotos.items():
                    edge_signature = (node_name, target)
                    if target and edge_signature not in analyzed_signatures:
                        origin_location = analyzer.goto_source_locations.get(target)
                        self.edges.append({
                             "source": node_name,
                             "target": target,
                             "condition": {
                                 "type": "exit_point" if target == "End" else "entry_point" if target == "Start" else "dynamic_goto"
                                #  "value": None,
                                #  "resolution": resolution_info
                             },
                             "metadata": {
                                 "origin_function": function_name,
                                 "origin_location": origin_location
                             }
                         })
                        analyzed_signatures.add(edge_signature)

                # if not analyzer.potential_gotos and analyzer.unresolved_reasons:
                #     edge_signature = (node_name, "?")
                #     if edge_signature not in analyzed_signatures:
                #          details = ", ".join(sorted([f"{t}:{d}" for t, d in analyzer.unresolved_reasons]))
                #          origin_location = None
                #          self.edges.append({
                #              "source": node_name, "target": "?",
                #              "condition": {
                #                 "type": "unknown"
                #                 #  "value": None,
                #                 #  "resolution": {"status": "unresolved", "detail": details}
                #              },
                #              "metadata": {"origin_function": function_name, "origin_location": origin_location, "unresolved": True}
                #          })
                #          analyzed_signatures.add(edge_signature)
            else:
                if node_name not in ["Start", "End"]:
                    print(f"  Warning: Function '{function_name}' mapped to node '{node_name}' not found in definitions.")


def extract_langgraph_graph(directory_path: str, output_filename: str):
    visitor = GraphVisitor()
    print(f"Starting graph extraction in directory: {directory_path}")

    if not os.path.isdir(directory_path):
         print(f"Error: Provided path '{directory_path}' is not a valid directory.", file=sys.stderr)
         sys.exit(1)

    all_files_processed = []
    parse_errors = 0

    for root, dirs, files in os.walk(directory_path):
        dirs[:] = [d for d in dirs if not d.startswith('.') and d not in ['venv', 'env', '__pycache__', 'node_modules', '.git']]

        for filename in files:
            if filename.endswith(".py"):
                filepath = os.path.join(root, filename)
                all_files_processed.append(filepath)
                print(f"Processing file: {filepath}")
                try:
                    with open(filepath, "r", encoding='utf-8') as f: content = f.read()
                    tree = ast.parse(content, filename=filepath)

                    for node in ast.walk(tree):
                         if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                             node._filename = filepath 

                    visitor.visit_file(tree, filepath) 

                except SyntaxError as e:
                    print(f"Warning: Skipping file {filepath} due to SyntaxError: {e}", file=sys.stderr)
                    parse_errors += 1
                except Exception as e:
                    print(f"Warning: Skipping file {filepath} due to unexpected error: {e}", file=sys.stderr)
                    parse_errors += 1

    try:
        visitor.analyze_functions_for_goto()
    except Exception as e:
        print(f"Error during post-processing analysis: {e}", file=sys.stderr)
        sys.exit(1)


    graph_data = {"nodes": visitor.nodes, "edges": visitor.edges}
    print(f"Finished graph extraction. Found {len(visitor.nodes)} nodes and {len(visitor.edges)} edges across {len(all_files_processed)} files ({parse_errors} file processing errors).")


    output_path = Path(output_filename)
    try:
        output_path.parent.mkdir(parents=True, exist_ok=True)

        graph_data["nodes"].sort(key=lambda x: x.get("id", ""))
        graph_data["edges"].sort(key=lambda x: (x.get("source", ""), x.get("target", "")))

        with open(output_path, "w", encoding='utf-8') as outfile:
            json.dump(graph_data, outfile, indent=2) 
        print(f"Combined graph data written to {output_path}")

    except Exception as e:
        print(f"Error writing JSON output to {output_path}: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Extract LangGraph structures from Python code into a graph JSON.")
    parser.add_argument("directory", type=str, nargs='?', default=".", help="Directory containing the Python code (default: current directory).")
    parser.add_argument("-o", "--output", type=str, default="langgraph_graph_output.json", help="Output JSON filename (default: langgraph_graph_output.json).")
    args = parser.parse_args()

    target_directory = args.directory
    output_filename = args.output

    extract_langgraph_graph(target_directory, output_filename)


import argparse
import ast
import json
import os
from pathlib import Path
import sys
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from enum import Enum
import yaml


def get_potential_fqn(node: ast.AST) -> Optional[str]:
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        base = get_potential_fqn(node.value)
        return f"{base}.{node.attr}" if base else node.attr
    return None

def _safe_load_yaml(filepath: str) -> Optional[Dict]:
    if not os.path.exists(filepath):
        return None
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            return yaml.safe_load(f)
    except yaml.YAMLError as e:
        print(f"Warning: Error parsing YAML file {filepath}: {e}", file=sys.stderr)
    except IOError as e:
        print(f"Warning: Error reading YAML file {filepath}: {e}", file=sys.stderr)
    except Exception as e:
        print(f"Warning: Unexpected error loading YAML {filepath}: {e}", file=sys.stderr)
    return None

def _stringify_ast_node(node: ast.AST) -> str:
    if isinstance(node, ast.Constant): return repr(node.value)
    elif isinstance(node, ast.Name): return node.id
    elif isinstance(node, ast.Attribute): return get_potential_fqn(node) or f"<Attribute {node.attr}>"
    elif isinstance(node, ast.Subscript):
        base = _stringify_ast_node(node.value)
        slc = _stringify_ast_node(node.slice)
        return f"{base}[{slc}]"
    elif isinstance(node, ast.Index): # support for Python < 3.9
         return _stringify_ast_node(node.value)
    elif isinstance(node, ast.Tuple):
         return f"({', '.join(_stringify_ast_node(elt) for elt in node.elts)})"
    elif isinstance(node, ast.List):
         return f"[{', '.join(_stringify_ast_node(elt) for elt in node.elts)}]"
    elif isinstance(node, ast.Dict):
         pairs = [f"{_stringify_ast_node(k)}: {_stringify_ast_node(v)}" for k, v in zip(node.keys, node.values)]
         return f"{{{', '.join(pairs)}}}"
    elif isinstance(node, ast.Call):
         func = _stringify_ast_node(node.func)
         args = [_stringify_ast_node(arg) for arg in node.args]
         kwargs = [f"{kw.arg}={_stringify_ast_node(kw.value)}" for kw in node.keywords if kw.arg]
         return f"{func}({', '.join(args + kwargs)})"
    else:
        try: return ast.dump(node)
        except Exception: return f"<AST:{type(node).__name__}>"


class CrewProcess(str, Enum):
    SEQUENTIAL = "sequential"
    HIERARCHICAL = "hierarchical"
    UNKNOWN = "unknown"


NODE_TYPE_MAP = {
    "agent": "Agent",
    "task": "Generic",
    "tool": "ToolExecutor",
    "custom_tool": "ToolExecutor",
    "start": "Special",
    "end": "Special",
}


class CrewAIStructureExtractor(ast.NodeVisitor):
    def __init__(self):
        self.agents_info: Dict[str, Dict[str, Any]] = {}
        self.tasks_info: Dict[str, Dict[str, Any]] = {}
        self.crews_info: Dict[str, Dict[str, Any]] = {}
        self.tools_info: Dict[str, Dict[str, Any]] = {}
        self.task_lists: Dict[str, Dict[str, Any]] = {}
        self.agent_lists: Dict[str, Dict[str, Any]] = {}
        # --- New: Track self.attribute assignments for tool resolution ---
        self.tool_instance_map: Dict[str, Dict[str, ast.AST]] = defaultdict(dict) # class_name -> {attr_name: value_node}

        self.imports: Dict[str, str] = {}
        self.tool_imports: Set[str] = set()
        self.current_filepath: Optional[str] = None
        self.current_class_name: Optional[str] = None
        self.current_crew_base_info: Optional[Dict] = None

        self.crewai_fqns = {
            "Agent": "crewai.agent.Agent",
            "Task": "crewai.task.Task",
            "Crew": "crewai.crew.Crew",
            "Process": "crewai.process.Process",
            "CrewBase": "crewai.project.crew_base.CrewBase",
            "agent": "crewai.project.decorators.agent",
            "task": "crewai.project.decorators.task",
            "crew": "crewai.project.decorators.crew",
            "tool": "crewai.tools.tool.tool",
            "BaseTool": "crewai_tools.tools.base_tool.BaseTool"
        }

        self.output_nodes: List[Dict[str, Any]] = []
        self.output_edges: List[Dict[str, Any]] = []
        self.output_node_names: Set[str] = set()


    def _get_location_dict(self, node: ast.AST) -> Optional[Dict[str, Any]]:
        if not self.current_filepath or not hasattr(node, 'lineno'):
             return None
        return {
            "file": self.current_filepath,
            "line": node.lineno,
            "col": node.col_offset,
            "end_line": getattr(node, 'end_lineno', node.lineno),
            "end_col": getattr(node, 'end_col_offset', -1),
        }

    def _add_output_node(self, name: str, function_name: Optional[str], docstring: Optional[str],
                         node_type: str, source_location: Optional[Dict], metadata: Optional[Dict]):
        if name not in self.output_node_names:
            node_data = {
                "name": name,
                "function_name": function_name,
                "docstring": docstring,
                "node_type": node_type,
                "source_location": source_location,
                "metadata": metadata or {}
            }
            self.output_nodes.append(node_data)
            self.output_node_names.add(name)

    def _add_output_edge(self, source: str, target: str, condition: Dict, metadata: Optional[Dict]):
        if source in self.output_node_names and target in self.output_node_names:
            # TODO: Add more robust duplicate check if needed
            self.output_edges.append({
                "source": source,
                "target": target,
                "condition": condition,
                "metadata": metadata or {}
            })

    def _resolve_fqn_from_node(self, node: ast.AST) -> Optional[str]:
        potential_name = get_potential_fqn(node)
        if potential_name:
             if potential_name in self.imports:
                  return self.imports[potential_name]
             parts = potential_name.split('.')
             if len(parts) > 1:
                  base_alias = parts[0]
                  if base_alias in self.imports:
                       if self.imports[base_alias] == base_alias:
                            return potential_name
                       else:
                            return f"{self.imports[base_alias]}.{'.'.join(parts[1:])}"
             return potential_name
        return None

    def _get_string_value(self, node: ast.AST) -> Optional[str]:
        if isinstance(node, ast.Constant) and isinstance(node.value, str):
            return node.value
        elif isinstance(node, ast.JoinedStr):
             try:
                  return "".join(str(v.value) for v in node.values if isinstance(v, ast.Constant))
             except: return "<f-string>"
        return None

    def _get_list_of_identifiers(self, node: ast.AST) -> Optional[List[str]]:
        if isinstance(node, ast.List):
            elements = []
            for elt in node.elts:
                name = get_potential_fqn(elt)
                if name:
                    elements.append(name)
                elif isinstance(elt, ast.Call):
                     call_name = get_potential_fqn(elt.func)
                     if call_name:
                          elements.append(call_name)
                else:
                    elements.append(_stringify_ast_node(elt))
            return elements
        return None

    def _get_call_kwargs(self, node: ast.Call) -> Dict[str, Any]:
        kwargs = {}
        for keyword in node.keywords:
            arg_name = keyword.arg
            if arg_name:
                value_node = keyword.value
                if isinstance(value_node, ast.Constant):
                    kwargs[arg_name] = value_node.value
                elif isinstance(value_node, (ast.Name, ast.Attribute)):
                    resolved_name = get_potential_fqn(value_node)
                    kwargs[arg_name] = resolved_name if resolved_name else _stringify_ast_node(value_node)
                elif isinstance(value_node, ast.List):
                    kwargs[arg_name] = self._get_list_of_identifiers(value_node) or _stringify_ast_node(value_node)
                else:
                    kwargs[arg_name] = _stringify_ast_node(value_node)
        return kwargs

    def _parse_config_arg(self, node: ast.AST, base_dir: str) -> Tuple[Optional[str], Optional[Dict]]:
        config_ref_str = _stringify_ast_node(node)
        config_data = None
        if isinstance(node, ast.Constant) and isinstance(node.value, str):
            config_path = node.value
            abs_path = os.path.join(base_dir, config_path) if not os.path.isabs(config_path) else config_path
            config_data = _safe_load_yaml(abs_path)
        return config_ref_str, config_data


    def _find_return_call(self, body_nodes: List[ast.AST], target_class_fqn: str) -> Optional[ast.Call]:
        for stmt in body_nodes:
            if isinstance(stmt, ast.Return) and isinstance(stmt.value, ast.Call):
                call_node = stmt.value
                func_fqn = self._resolve_fqn_from_node(call_node.func)
                func_name_simple = get_potential_fqn(call_node.func)
                if func_fqn == target_class_fqn or func_name_simple == target_class_fqn or \
                   (isinstance(call_node.func, ast.Name) and target_class_fqn.endswith(f".{call_node.func.id}")):
                    return call_node
            elif isinstance(stmt, (ast.If, ast.Try)): # Recurse into simple blocks
                call = self._find_return_call(stmt.body, target_class_fqn)
                if call: return call
                if hasattr(stmt, 'orelse'):
                    call = self._find_return_call(stmt.orelse, target_class_fqn)
                    if call: return call
                if hasattr(stmt, 'finalbody'):
                     call = self._find_return_call(stmt.finalbody, target_class_fqn)
                     if call: return call
            # TODO: Could add recursion for If/Try blocks if needed (more complex)
        return None


    def visit(self, node: ast.AST, filepath: Optional[str] = None):
        original_path = self.current_filepath
        if filepath:
            self.current_filepath = filepath
        super().visit(node)
        self.current_filepath = original_path

    def visit_Import(self, node: ast.Import):
        for alias in node.names:
             module_name = alias.name
             local_name = alias.asname or module_name
             self.imports[local_name] = module_name
             if module_name == 'crewai':
                  self.crewai_fqns['Agent'] = 'crewai.Agent'
                  self.crewai_fqns['Task'] = 'crewai.Task'
                  self.crewai_fqns['Crew'] = 'crewai.Crew'
                  self.crewai_fqns['Process'] = 'crewai.Process'
                  self.crewai_fqns['CrewBase'] = 'crewai.CrewBase'
                  self.crewai_fqns['agent'] = 'crewai.agent'
                  self.crewai_fqns['task'] = 'crewai.task'
                  self.crewai_fqns['crew'] = 'crewai.crew'
             if module_name == 'crewai_tools':
                  self.crewai_fqns['BaseTool'] = 'crewai_tools.BaseTool'
                  self.crewai_fqns['tool'] = 'crewai_tools.tool'

    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node.module:
            base_module = node.module
            for alias in node.names:
                imported_name = alias.name
                local_name = alias.asname or imported_name
                full_path = f"{base_module}.{imported_name}"
                self.imports[local_name] = full_path

                for key, default_fqn in self.crewai_fqns.items():
                    if full_path == default_fqn:
                        self.crewai_fqns[key] = local_name
                        break

                if base_module == 'crewai_tools' or base_module.startswith('crewai_tools.'):
                    self.tool_imports.add(imported_name)
                    if alias.asname:
                         self.tool_imports.add(local_name)

    def visit_ClassDef(self, node: ast.ClassDef):
         original_class_name = self.current_class_name
         original_crew_base_info = self.current_crew_base_info
         is_crew_base = False
         class_location = self._get_location_dict(node)
         class_docstring = ast.get_docstring(node)

         cb_decorator_name = self.crewai_fqns.get('CrewBase', 'crewai.project.CrewBase')
         for decorator in node.decorator_list:
              dec_name = get_potential_fqn(decorator)
              if dec_name == 'CrewBase' or self._resolve_fqn_from_node(decorator) == cb_decorator_name:
                   is_crew_base = True
                   break

         if is_crew_base:
              self.current_class_name = node.name
              self.current_crew_base_info = {"name": node.name, "agents": [], "tasks": [], "crew_method": None, "location": class_location, "docstring": class_docstring}

         bt_class_name = self.crewai_fqns.get('BaseTool', 'crewai_tools.BaseTool')
         is_custom_tool_class = False
         for base in node.bases:
             resolved_base_fqn = self._resolve_fqn_from_node(base)
             if resolved_base_fqn == bt_class_name or (isinstance(base, ast.Name) and base.id == 'BaseTool'):
                 is_custom_tool_class = True
                 break

         if is_custom_tool_class:
             tool_name = node.name
             tool_description = class_docstring or f"Custom tool class: {tool_name}"
             self.tools_info[tool_name] = {
                 "name": tool_name,
                 "function_name": tool_name,
                 "docstring": class_docstring,
                 "is_custom": True,
                 "description": tool_description,
                 "var_name": None,
                 "location": class_location,
                 "type": "class"
             }

         self.generic_visit(node)

         # --- Populate tool_instance_map after visiting class body (including __init__) ---
         if self.current_class_name == node.name: # Check if we're back from visiting this class
            init_method = next((n for n in node.body if isinstance(n, ast.FunctionDef) and n.name == '__init__'), None)
            if init_method:
                for stmt in init_method.body:
                    if isinstance(stmt, ast.Assign) and len(stmt.targets) == 1:
                        target = stmt.targets[0]
                        if isinstance(target, ast.Attribute) and isinstance(target.value, ast.Name) and target.value.id == 'self':
                            attr_name = target.attr
                            # Check if the assigned value is a tool instance we care about
                            value_node = stmt.value
                            tool_class_name = None
                            if isinstance(value_node, ast.Call):
                                tool_class_name = get_potential_fqn(value_node.func)

                            # Only store if it looks like a known tool or custom tool
                            if tool_class_name and (tool_class_name in self.tool_imports or tool_class_name in self.tools_info):
                                self.tool_instance_map[node.name][attr_name] = value_node


         self.current_class_name = original_class_name
         self.current_crew_base_info = original_crew_base_info


    def visit_FunctionDef(self, node: ast.FunctionDef):
        func_name = node.name
        is_crewai_decorated = False
        base_dir = os.path.dirname(self.current_filepath) if self.current_filepath else "."
        func_location = self._get_location_dict(node)
        func_docstring = ast.get_docstring(node)

        agent_decorator = self.crewai_fqns.get('agent')
        task_decorator = self.crewai_fqns.get('task')
        crew_decorator = self.crewai_fqns.get('crew')
        tool_decorator = self.crewai_fqns.get('tool')

        for decorator in node.decorator_list:
            decorator_fqn = self._resolve_fqn_from_node(decorator)
            decorator_name = get_potential_fqn(decorator)

            if self.current_crew_base_info and (decorator_fqn == agent_decorator or decorator_name == 'agent'):
                 is_crewai_decorated = True
                 agent_call = self._find_return_call(node.body, self.crewai_fqns['Agent'])
                 if agent_call:
                      agent_id = func_name
                      kwargs = self._get_call_kwargs(agent_call)
                      config_ref, config_data = (None, None)
                      config_kw_node = next((kw.value for kw in agent_call.keywords if kw.arg == 'config'), None)
                      if config_kw_node:
                           config_ref, config_data = self._parse_config_arg(config_kw_node, base_dir)

                      tool_refs = kwargs.get("tools", [])
                      self.agents_info[agent_id] = {
                           "name": agent_id,
                           "function_name": agent_id,
                           "docstring": func_docstring,
                           "location": func_location,
                           "config_ref": config_ref,
                           "config_data": config_data,
                           "tools_vars": tool_refs if isinstance(tool_refs, list) else [],
                           "metadata_kwargs": kwargs,
                           "is_decorator": True,
                           "class_context": self.current_class_name
                      }
                      self.current_crew_base_info["agents"].append(agent_id)
                 break

            elif self.current_crew_base_info and (decorator_fqn == task_decorator or decorator_name == 'task'):
                 is_crewai_decorated = True
                 task_call = self._find_return_call(node.body, self.crewai_fqns['Task'])
                 if task_call:
                      task_id = func_name
                      kwargs = self._get_call_kwargs(task_call)
                      config_ref, config_data = (None, None)
                      config_kw_node = next((kw.value for kw in task_call.keywords if kw.arg == 'config'), None)
                      if config_kw_node:
                           config_ref, config_data = self._parse_config_arg(config_kw_node, base_dir)

                      agent_var = None # Explicitly set to None for decorator tasks initially
                      agent_kw_node = next((kw.value for kw in task_call.keywords if kw.arg == 'agent'), None)
                      if agent_kw_node: # If agent= IS present, parse it
                         agent_ref_node_str = _stringify_ast_node(agent_kw_node)
                         if isinstance(agent_ref_node_str, str):
                            agent_var = agent_ref_node_str.removesuffix('()')
                            if agent_var.startswith('self.'):
                                agent_var = agent_var[5:]

                      output_ref = kwargs.get("output_json") or kwargs.get("output_pydantic") or kwargs.get("output_file")
                      context_vars = kwargs.get("context", [])
                      # --- Read dependencies from YAML if config was loaded ---
                      dependencies = []
                      if config_data and isinstance(config_data.get('dependencies'), list):
                          dependencies = config_data['dependencies']
                      # --- Also check kwargs for dependencies (less common) ---
                      elif isinstance(kwargs.get('dependencies'), list):
                           dependencies = kwargs['dependencies']

                      self.tasks_info[task_id] = {
                           "name": task_id,
                           "function_name": task_id,
                           "docstring": func_docstring,
                           "location": func_location,
                           "config_ref": config_ref,
                           "config_data": config_data,
                           "agent_var": agent_var,
                           "description": kwargs.get("description", config_data.get("description", task_id) if config_data else task_id),
                           "context_vars": context_vars if isinstance(context_vars, list) else [],
                           "output_ref": output_ref,
                           "metadata_kwargs": kwargs,
                           "dependencies": dependencies, # Store dependencies
                           "is_decorator": True,
                           "class_context": self.current_class_name
                      }
                      self.current_crew_base_info["tasks"].append(task_id)
                 break

            elif self.current_crew_base_info and (decorator_fqn == crew_decorator or decorator_name == 'crew'):
                is_crewai_decorated = True
                crew_call = self._find_return_call(node.body, self.crewai_fqns['Crew'])
                if crew_call:
                     crew_id = self.current_class_name
                     kwargs = self._get_call_kwargs(crew_call)

                     agents_list_ref = kwargs.get("agents")
                     if agents_list_ref == "self.agents":
                          agents_list_ref = self.current_crew_base_info["agents"]

                     tasks_list_ref = kwargs.get("tasks")
                     if tasks_list_ref == "self.tasks":
                          tasks_list_ref = self.current_crew_base_info["tasks"]

                     process_type_str = kwargs.get("process", "sequential")
                     process_type = self._parse_process_type(process_type_str)

                     self.crews_info[crew_id] = {
                          "name": crew_id,
                          "location": self.current_crew_base_info.get("location"),
                          "definition_location": func_location,
                          "agents_list_ref": agents_list_ref,
                          "tasks_list_ref": tasks_list_ref,
                          "process_type": process_type,
                          "metadata_kwargs": kwargs,
                          "is_decorator": True,
                          "class_context": self.current_class_name
                     }
                     self.current_crew_base_info["crew_method"] = func_name
                break

            elif decorator_fqn == tool_decorator or decorator_name == 'tool' or (isinstance(decorator, ast.Call) and get_potential_fqn(decorator.func) == 'tool'):
                is_crewai_decorated = True
                tool_name = func_name
                tool_description = func_docstring or f"Custom tool function: {tool_name}"
                self.tools_info[tool_name] = {
                     "name": tool_name,
                     "function_name": tool_name,
                     "docstring": func_docstring,
                     "location": func_location,
                     "is_custom": True,
                     "description": tool_description,
                     "var_name": None,
                     "type": "function"
                }
                break

        if not is_crewai_decorated:
            self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign):
        assign_location = self._get_location_dict(node)
        if len(node.targets) == 1 :
            target_node = node.targets[0]

            # --- Handle self.attribute assignment ---
            if isinstance(target_node, ast.Attribute) and isinstance(target_node.value, ast.Name) and target_node.value.id == 'self' and self.current_class_name:
                attr_name = target_node.attr
                value_node = node.value
                tool_class_name = None
                if isinstance(value_node, ast.Call):
                    tool_class_name = get_potential_fqn(value_node.func)

                if tool_class_name and (tool_class_name in self.tool_imports or tool_class_name in self.tools_info):
                     self.tool_instance_map[self.current_class_name][attr_name] = value_node
            # --- End self.attribute handling ---

            elif isinstance(target_node, ast.Name): # Handle regular variable assignment
                target_var = target_node.id
                value_node = node.value
                base_dir = os.path.dirname(self.current_filepath) if self.current_filepath else "."

                if isinstance(value_node, ast.List):
                    ids = self._get_list_of_identifiers(value_node)
                    if ids:
                        list_info = {"tasks": ids, "agents": ids, "location": assign_location}
                        self.task_lists[target_var] = list_info
                        self.agent_lists[target_var] = list_info

                elif isinstance(value_node, ast.Call):
                    call_func = value_node.func
                    kwargs = self._get_call_kwargs(value_node)
                    func_fqn = self._resolve_fqn_from_node(call_func)
                    func_name_simple = get_potential_fqn(call_func)

                    agent_fqn = self.crewai_fqns['Agent']
                    task_fqn = self.crewai_fqns['Task']
                    crew_fqn = self.crewai_fqns['Crew']

                    if func_fqn == agent_fqn or func_name_simple == 'Agent':
                        config_ref, config_data = (None, None)
                        config_kw_node = next((kw.value for kw in value_node.keywords if kw.arg == 'config'), None)
                        if config_kw_node: config_ref, config_data = self._parse_config_arg(config_kw_node, base_dir)
                        tool_refs = kwargs.get("tools", [])
                        self.agents_info[target_var] = {
                            "name": target_var,
                            "function_name": target_var,
                            "docstring": None,
                            "location": assign_location,
                            "config_ref": config_ref, "config_data": config_data,
                            "tools_vars": tool_refs if isinstance(tool_refs, list) else [],
                            "metadata_kwargs": kwargs, "is_decorator": False,
                            "class_context": self.current_class_name
                        }

                    elif func_fqn == task_fqn or func_name_simple == 'Task':
                        config_ref, config_data = (None, None)
                        config_kw_node = next((kw.value for kw in value_node.keywords if kw.arg == 'config'), None)
                        if config_kw_node: config_ref, config_data = self._parse_config_arg(config_kw_node, base_dir)

                        agent_var_raw = kwargs.get("agent")
                        agent_var = agent_var_raw if isinstance(agent_var_raw, str) else None

                        output_ref = kwargs.get("output_json") or kwargs.get("output_pydantic") or kwargs.get("output_file")
                        context_vars = kwargs.get("context", [])
                        dependencies = kwargs.get("dependencies", []) # Check kwargs too

                        self.tasks_info[target_var] = {
                            "name": target_var,
                            "function_name": target_var,
                            "docstring": None,
                            "location": assign_location,
                            "config_ref": config_ref, "config_data": config_data,
                            "agent_var": agent_var,
                            "description": kwargs.get("description", target_var),
                            "context_vars": context_vars if isinstance(context_vars, list) else [],
                            "output_ref": output_ref, "metadata_kwargs": kwargs,
                            "dependencies": dependencies if isinstance(dependencies, list) else [],
                            "is_decorator": False, "class_context": self.current_class_name
                        }

                    elif func_fqn == crew_fqn or func_name_simple == 'Crew':
                        tasks_list_ref = kwargs.get("tasks")
                        agents_list_ref = kwargs.get("agents")
                        process_type_str = kwargs.get("process", "sequential")
                        process_type = self._parse_process_type(process_type_str)
                        self.crews_info[target_var] = {
                            "name": target_var,
                            "location": assign_location,
                            "definition_location": assign_location,
                            "agents_list_ref": agents_list_ref,
                            "tasks_list_ref": tasks_list_ref, "process_type": process_type,
                            "metadata_kwargs": kwargs, "is_decorator": False,
                            "class_context": self.current_class_name
                        }

                    else:
                         tool_class_name = get_potential_fqn(call_func)
                         if tool_class_name and tool_class_name in self.tool_imports:
                              if tool_class_name in self.tools_info:
                                   self.tools_info[tool_class_name]["var_name"] = target_var
                                   self.tools_info[tool_class_name]["assign_location"] = assign_location
                                   self.tools_info[target_var] = self.tools_info[tool_class_name]
                              else:
                                   self.tools_info[target_var] = {
                                        "name": target_var,
                                        "function_name": tool_class_name,
                                        "docstring": None,
                                        "location": assign_location,
                                        "is_custom": False,
                                        "description": f"Predefined tool instance: {tool_class_name}",
                                        "var_name": target_var,
                                        "tool_class": tool_class_name,
                                        "type": "instance"
                                   }


        self.generic_visit(node)

    def _parse_process_type(self, process_ref: Any) -> CrewProcess:
        proc_seq_fqn = self.crewai_fqns['Process'] + ".sequential"
        proc_hier_fqn = self.crewai_fqns['Process'] + ".hierarchical"
        process_ref_str = str(process_ref)

        if process_ref_str == proc_seq_fqn or process_ref_str.endswith('.sequential') or process_ref_str == 'sequential':
            return CrewProcess.SEQUENTIAL
        elif process_ref_str == proc_hier_fqn or process_ref_str.endswith('.hierarchical') or process_ref_str == 'hierarchical':
            return CrewProcess.HIERARCHICAL
        return CrewProcess.UNKNOWN


    def finalize_graph(self):
        print("Finalizing graph...")

        self._add_output_node(name="Start", function_name=None, docstring=None,
                             node_type=NODE_TYPE_MAP["start"], source_location=None, metadata={})
        self._add_output_node(name="End", function_name=None, docstring=None,
                             node_type=NODE_TYPE_MAP["end"], source_location=None, metadata={})

        for agent_id, agent_data in self.agents_info.items():
             metadata = agent_data.get("metadata_kwargs", {})
             if agent_data.get("config_data"):
                  metadata['role'] = agent_data["config_data"].get('role', metadata.get('role'))
                  metadata['goal'] = agent_data["config_data"].get('goal', metadata.get('goal'))
                  metadata['backstory'] = agent_data["config_data"].get('backstory', metadata.get('backstory'))
             metadata['config_ref'] = agent_data.get('config_ref')
             metadata['config_resolved'] = bool(agent_data.get("config_data"))
             metadata['tools'] = agent_data.get('tools_vars')
             metadata['is_decorator'] = agent_data.get('is_decorator')
             metadata['class_context'] = agent_data.get('class_context')

             self._add_output_node(name=agent_id,
                                  function_name=agent_data.get("function_name"),
                                  docstring=agent_data.get("docstring"),
                                  node_type=NODE_TYPE_MAP["agent"],
                                  source_location=agent_data.get("location"),
                                  metadata=metadata)

        for task_id, task_data in self.tasks_info.items():
             metadata = task_data.get("metadata_kwargs", {})
             if task_data.get("config_data"):
                  metadata['description'] = task_data["config_data"].get('description', metadata.get('description'))
             metadata['config_ref'] = task_data.get('config_ref')
             metadata['config_resolved'] = bool(task_data.get("config_data"))
             metadata['agent'] = task_data.get('agent_var') # Keep the potentially null agent_var here
             metadata['context'] = task_data.get('context_vars')
             metadata['output_ref'] = task_data.get('output_ref')
             metadata['is_decorator'] = task_data.get('is_decorator')
             metadata['class_context'] = task_data.get('class_context')
             metadata['dependencies'] = task_data.get('dependencies') # Add dependencies to metadata
             node_description = metadata.get('description') or task_data.get('description')

             self._add_output_node(name=task_id,
                                  function_name=task_data.get("function_name"),
                                  docstring=task_data.get("docstring"),
                                  node_type=NODE_TYPE_MAP["task"],
                                  source_location=task_data.get("location"),
                                  metadata=metadata)

        tool_node_map = {}
        final_tool_nodes = {}

        # --- Pass 1: Process explicitly defined/instantiated tools ---
        for tool_ref, tool_data in self.tools_info.items():
             node_name = tool_data.get("var_name") or tool_data["name"]
             if node_name not in final_tool_nodes:
                  metadata = {
                       "description": tool_data["description"],
                       "is_custom": tool_data["is_custom"],
                       "definition_type": tool_data.get("type"),
                       "tool_reference": tool_data["name"]
                  }
                  final_tool_nodes[node_name] = {
                      "name": node_name,
                      "function_name": tool_data["function_name"],
                      "docstring": tool_data.get("docstring"),
                      "node_type": NODE_TYPE_MAP["tool"],
                      "source_location": tool_data.get("location") or tool_data.get("assign_location"),
                      "metadata": metadata
                  }
             # Map original name and var_name (if exists) to the final node name
             tool_node_map[tool_data["name"]] = node_name
             if tool_data.get("var_name"):
                 tool_node_map[tool_data["var_name"]] = node_name

        # --- Pass 2: Resolve self. references and create implicit nodes ---
        all_tool_refs_found = set()
        for agent_data in self.agents_info.values():
            all_tool_refs_found.update(agent_data.get('tools_vars', []))

        for tool_ref in all_tool_refs_found:
            if tool_ref in tool_node_map: continue # Already handled

            resolved_node_name = None
            # Try resolving self.attribute
            if isinstance(tool_ref, str) and tool_ref.startswith('self.'):
                class_context = agent_data.get('class_context') # Assumes agent_data is in scope (needs adjustment if run outside agent loop)
                attr_name = tool_ref[5:]
                if class_context and attr_name in self.tool_instance_map.get(class_context, {}):
                    value_node = self.tool_instance_map[class_context][attr_name]
                    # Find the original tool class/function name from the assignment
                    tool_class_name = None
                    if isinstance(value_node, ast.Call):
                        tool_class_name = get_potential_fqn(value_node.func)
                    elif isinstance(value_node, ast.Name): # Assigned from another var/tool func
                        tool_class_name = value_node.id
                    # Map this class/func name back to a node name if possible
                    if tool_class_name:
                        resolved_node_name = tool_node_map.get(tool_class_name)

            # If not resolved via self. or already mapped, check if it's an imported tool name
            if not resolved_node_name and tool_ref in self.tool_imports and tool_ref not in tool_node_map:
                 resolved_node_name = tool_ref # Use the import name directly

            # If still not resolved, create implicit node
            if not resolved_node_name:
                 node_name = tool_ref
                 if node_name not in final_tool_nodes:
                    metadata = {
                        "description": f"Implicitly referenced tool: {tool_ref}",
                        "is_custom": False,
                        "definition_type": "reference",
                        "tool_reference": tool_ref
                    }
                    final_tool_nodes[node_name] = {
                        "name": node_name, "function_name": None, "docstring": None,
                        "node_type": NODE_TYPE_MAP["tool"], "source_location": None,
                        "metadata": metadata
                    }
                 resolved_node_name = node_name

            # Map the original reference (e.g., 'self.whisper_tool') to the final node name
            if resolved_node_name:
                tool_node_map[tool_ref] = resolved_node_name

        # Add all collected tool nodes
        for node_name, node_creation_data in final_tool_nodes.items():
             self._add_output_node(**node_creation_data)


        # --- Create Edges ---
        # 5. Agent -> Task Assignment (Handles decorator inference)
        for task_id, task_data in self.tasks_info.items():
            agent_node_name = task_data.get("agent_var") # Explicitly assigned agent

            # If decorator task and no agent explicitly assigned in Task(), infer by name matching
            if task_data.get("is_decorator") and not agent_node_name:
                task_name = task_data["name"]
                class_context = task_data.get("class_context")
                # Check if an agent with the same name exists in the same class context
                if class_context and task_name in self.agents_info and self.agents_info[task_name].get("class_context") == class_context:
                    agent_node_name = task_name # Inferred agent name

            if agent_node_name and agent_node_name in self.agents_info:
                 condition = {"type": "assignment"}
                 metadata = {"definition_location": task_data.get("location")}
                 self._add_output_edge(agent_node_name, task_id, condition, metadata)
            # else:
            #      print(f"Debug: Could not link agent to task '{task_id}'. Explicit: '{task_data.get('agent_var')}', Inferred: '{agent_node_name}'")


        # 6. Agent -> Tool Usage (Uses resolved tool_node_map)
        for agent_id, agent_data in self.agents_info.items():
             tool_refs = agent_data.get("tools_vars", [])
             agent_def_location = agent_data.get("location")
             for tool_ref in tool_refs:
                 tool_node_name = tool_node_map.get(tool_ref) # Lookup resolved name
                 if tool_node_name:
                      condition = {"type": "tool_usage"}
                      metadata = {"definition_location": agent_def_location}
                      self._add_output_edge(agent_id, tool_node_name, condition, metadata)
                 # else:
                 #      print(f"Debug: Could not find tool node for reference '{tool_ref}' used by agent '{agent_id}'")


        # 7. Crew Execution Flow (Task sequencing)
        processed_tasks_in_crews = set()
        for crew_id, crew_data in self.crews_info.items():
            crew_def_location = crew_data.get("definition_location")
            task_sequence_refs = crew_data.get("tasks_list_ref")
            actual_task_sequence = []

            if isinstance(task_sequence_refs, list):
                 actual_task_sequence = task_sequence_refs
            elif isinstance(task_sequence_refs, str):
                 if task_sequence_refs in self.task_lists:
                      actual_task_sequence = self.task_lists[task_sequence_refs].get("tasks", [])
                      if not crew_def_location: crew_def_location = self.task_lists[task_sequence_refs].get("location")

            if not actual_task_sequence:
                 print(f"Warning: No valid task sequence found for crew '{crew_id}'.", file=sys.stderr)
                 continue

            valid_task_node_names = [t for t in actual_task_sequence if t in self.tasks_info]
            processed_tasks_in_crews.update(valid_task_node_names)

            if not valid_task_node_names: continue

            process_type = crew_data.get("process_type", CrewProcess.SEQUENTIAL)
            crew_metadata = {"crew_name": crew_id, "process_type": process_type.value, "definition_location": crew_def_location}

            if process_type == CrewProcess.SEQUENTIAL:
                start_condition = {"type": "entry_point"}
                self._add_output_edge("Start", valid_task_node_names[0], start_condition, crew_metadata)
                seq_condition = {"type": "static"}
                for i in range(len(valid_task_node_names) - 1):
                     self._add_output_edge(valid_task_node_names[i], valid_task_node_names[i+1], seq_condition, crew_metadata)
                end_condition = {"type": "finish_point"}
                self._add_output_edge(valid_task_node_names[-1], "End", end_condition, crew_metadata)

            elif process_type == CrewProcess.HIERARCHICAL:
                 start_condition = {"type": "entry_point", "detail": "hierarchical"}
                 end_condition = {"type": "finish_point", "detail": "hierarchical"}
                 for task_node_name in valid_task_node_names:
                     self._add_output_edge("Start", task_node_name, start_condition, crew_metadata)
                     self._add_output_edge(task_node_name, "End", end_condition, crew_metadata)
                 # TODO: Optionally add all-to-all task edges if needed for visualization

        # 8. Task -> Task (Context & Dependencies)
        for task_id, task_data in self.tasks_info.items():
             task_def_location = task_data.get("location")
             # Context Edges
             context_vars = task_data.get("context_vars", [])
             if isinstance(context_vars, list):
                 for ctx_task_var in context_vars:
                     if ctx_task_var in self.tasks_info:
                         condition = {"type": "context_dependency"}
                         metadata = {"definition_location": task_def_location}
                         self._add_output_edge(ctx_task_var, task_id, condition, metadata)
            # Dependency Edges
             dependencies = task_data.get("dependencies", [])
             if isinstance(dependencies, list):
                 for dep_task_name in dependencies:
                     if dep_task_name in self.tasks_info:
                         condition = {"type": "dependency"} # Use a specific type
                         metadata = {"definition_location": task_def_location} # Dependency defined in target task
                         # Edge direction: Dependency -> Dependent Task
                         self._add_output_edge(dep_task_name, task_id, condition, metadata)


    def get_graph_data(self) -> Dict[str, List[Dict[str, Any]]]:
        return {"nodes": self.output_nodes, "edges": self.output_edges}

def extract_crewai_graph(directory_path: str, output_file: str):
    extractor = CrewAIStructureExtractor()
    print(f"Starting CrewAI structure extraction in: {directory_path}")

    if not os.path.isdir(directory_path):
         print(f"Error: Provided path '{directory_path}' is not a valid directory.", file=sys.stderr)
         sys.exit(1)

    parsed_files = 0
    for root, _, files in os.walk(directory_path):
        path_parts = Path(root).parts
        if any(part.startswith('.') or part in ['venv', 'env', '__pycache__', 'node_modules', '.git'] for part in path_parts):
            continue

        for filename in files:
            if filename.endswith(".py"):
                filepath = os.path.join(root, filename)
                try:
                    with open(filepath, "r", encoding='utf-8') as f:
                        content = f.read()
                    tree = ast.parse(content, filename=filepath)
                    extractor.visit(tree, filepath=filepath)
                    parsed_files += 1
                except SyntaxError as e:
                    print(f"Warning: Skipping file {filepath} due to SyntaxError: {e}", file=sys.stderr)
                except Exception as e:
                    print(f"Warning: Skipping file {filepath} due to error: {e}", file=sys.stderr)

    print(f"AST parsing complete ({parsed_files} files processed). Finalizing graph structure...")
    extractor.finalize_graph()
    graph_structure = extractor.get_graph_data()
    print(f"Extraction finished. Found {len(graph_structure['nodes'])} nodes and {len(graph_structure['edges'])} edges.")


    if graph_structure["nodes"] or graph_structure["edges"]:
        try:
            graph_structure["nodes"].sort(key=lambda x: (x.get("node_type", ""), x.get("name", "")))
            graph_structure["edges"].sort(key=lambda x: (x.get("source", ""), x.get("target", "")))

            output_path = Path(output_file)
            output_path.parent.mkdir(parents=True, exist_ok=True)

            with open(output_path, "w", encoding='utf-8') as f:
                json.dump(graph_structure, f, indent=2)
            print(f"Successfully wrote graph data to {output_path}")
        except IOError as e:
            print(f"Error writing output file {output_path}: {e}", file=sys.stderr)
            sys.exit(1)
        except Exception as e:
             print(f"An unexpected error occurred during JSON serialization or path creation: {e}", file=sys.stderr)
             sys.exit(1)
    else:
        print("No graph structure extracted. Output file not written.")
        try:
            output_path = Path(output_file)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            with open(output_path, "w", encoding='utf-8') as f:
                json.dump({"nodes": [], "edges": []}, f, indent=2)
            print(f"Empty graph structure written to {output_path}")
        except Exception as e:
            print(f"Error writing empty JSON output to {output_path}: {e}")
            sys.exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Extract CrewAI structures from Python code into a graph JSON.")
    parser.add_argument("directory", type=str, nargs='?', default=".", help="Directory containing the Python code (default: current directory).")
    parser.add_argument("-o", "--output", type=str, default="crewai_graph_output.json", help="Output JSON filename (default: crewai_graph_output.json).")
    args = parser.parse_args()

    target_dir = args.directory
    output_file = args.output

    extract_crewai_graph(target_dir, output_file)


import shutil
import json
import importlib.resources as pkg_resources
from pathlib import Path


def generate_visualization(json_path: str, open_browser: bool = False):
    import repello_agent_wiz.templates

    # Read framework name from JSON
    with open(json_path, "r") as f:
        graph = json.load(f)
        framework = graph.get("metadata", {}).get("framework", "unknown")

    output_dir = Path(f"{framework}_vis")
    output_dir.mkdir(exist_ok=True)

    # Copy assets
    assets_path = pkg_resources.files(repello_agent_wiz.templates).joinpath("assets")
    target_assets_path = output_dir / "assets"
    shutil.copytree(assets_path, target_assets_path, dirs_exist_ok=True)

    # Copy index.html and inject graph JSON
    index_template = pkg_resources.files(repello_agent_wiz.templates).joinpath("index.html")
    index_text = index_template.read_text(encoding="utf-8")

    json_string = json.dumps(graph, indent=2)
    index_filled = index_text.replace("const data = {};", f"const data = {json_string};")

    with open(output_dir / "index.html", "w") as f:
        f.write(index_filled)

    print(f"[✓] Visualization HTML generated at: {output_dir}/index.html")

    if open_browser:
        import webbrowser
        webbrowser.open(f"file://{(output_dir / 'index.html').resolve()}")




import ast
import json
import os
import argparse
from pathlib import Path
import sys
from typing import Any, Dict, List, Optional, Set, Tuple, Union
from enum import Enum

try:
    from pydantic import BaseModel, Field, ValidationError
except ImportError:
    print("Error: pydantic is required. Please install it using 'pip install pydantic'")
    exit(1)

class NodeType(str, Enum):
    AGENT = "Agent"
    TOOL = "Tool" 
    CUSTOM_TOOL = "CustomTool"
    PREDEFINED_TOOL = "PredefinedTool"
    START = "Start"
    END = "End" 

class ToolType(str, Enum): 
    DEFAULT = "Default"

class ToolDefinition(BaseModel):
    name: str
    custom: bool
    description: Optional[str] = None

class AgentDefinition(BaseModel):
    name: str 
    tools: list[ToolDefinition] = Field(default_factory=list)
    handoffs: list[str] = Field(default_factory=list) 
    variable_name: Optional[str] = None 
    source_location: Optional[Dict] = None 
    model_ref: Optional[str] = None 

class InvalidAgentDefinitionError(Exception): pass
class InvalidHandoffSpecError(Exception): pass
class ToolDefinitionNotFoundError(Exception): pass


def get_positional_arg_node(call_node: ast.Call, index: int) -> Optional[ast.AST]:
    if not isinstance(call_node, ast.Call): raise TypeError("Expected an ast.Call node")
    if index < 0: return None
    return call_node.args[index] if index < len(call_node.args) else None

def get_kwarg_node(call_node: ast.Call, keyword: str) -> Optional[ast.AST]:
    if not isinstance(call_node, ast.Call): raise TypeError("Expected an ast.Call node")
    for kw in call_node.keywords:
        if kw.arg == keyword: return kw.value
    return None

def is_name_or_attr(node: ast.AST) -> bool:
    if isinstance(node, ast.Name): return True
    if isinstance(node, ast.Attribute): return True
    if isinstance(node, ast.Subscript): 
        return is_name_or_attr(node.value)
    return False

def get_identifier_string(node: ast.AST) -> Optional[str]:
    if isinstance(node, ast.Name): return node.id
    elif isinstance(node, ast.Attribute):
        try:
            if hasattr(ast, 'unparse'): return ast.unparse(node)
            base = get_identifier_string(node.value)
            return f"{base}.{node.attr}" if base else node.attr
        except: return node.attr 
    elif isinstance(node, ast.Subscript): 
        return get_identifier_string(node.value)
    return None

def is_matching_call(node: ast.AST, target_names: Set[str]) -> bool:
    if not isinstance(node, ast.Call): return False
    func_node = node.func 
    base_name = get_identifier_string(func_node)
    if base_name and base_name in target_names:
        return True
    if isinstance(func_node, ast.Attribute) and func_node.attr in target_names:
         return True
    return False

def find_decorator_node(node: Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef], target_names: Set[str]) -> Optional[ast.AST]:
    for decorator in node.decorator_list:
        deco_node_to_check = decorator
        if isinstance(decorator, ast.Call):
            deco_node_to_check = decorator.func 
        deco_name = get_identifier_string(deco_node_to_check)
        if deco_name and deco_name in target_names:
            return decorator 
    return None

def extract_string_literal(node: Optional[ast.AST]) -> Optional[str]:
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        return node.value
    return None

def extract_kwarg_string(call_node: ast.Call, keyword: str) -> Optional[str]:
     value_node = get_kwarg_node(call_node, keyword)
     return extract_string_literal(value_node)

def get_qualified_name(node: Union[ast.Name, ast.Attribute, ast.Call, ast.Subscript]) -> str:
    if isinstance(node, ast.Name): return node.id
    elif isinstance(node, ast.Attribute):
        base = get_qualified_name(node.value)
        return f"{base}.{node.attr}" if base else node.attr
    elif isinstance(node, ast.Call): return get_qualified_name(node.func)
    elif isinstance(node, ast.Subscript): return get_qualified_name(node.value) 
    return ""

def represent_node(node: Optional[ast.AST]) -> str:
    if node is None: return "None"
    if isinstance(node, ast.Constant): return repr(node.value)
    if isinstance(node, ast.Name): return node.id
    if isinstance(node, ast.Attribute):
        name = get_qualified_name(node)
        return name if name else f"<Attribute {node.attr}>"
    if isinstance(node, ast.Subscript): 
        base = represent_node(node.value)
        slice_val = represent_node(node.slice)
        return f"{base}[{slice_val}]"
    if isinstance(node, ast.List): return f"[{', '.join(represent_node(elt) for elt in node.elts)}]"
    if isinstance(node, ast.Call):
         func_str = represent_node(node.func)
         args_str = ', '.join(represent_node(arg) for arg in node.args)
         kwargs_str = ', '.join(f"{kw.arg}={represent_node(kw.value)}" for kw in node.keywords if kw.arg)
         all_args = f"{args_str}{', ' if args_str and kwargs_str else ''}{kwargs_str}"
         return f"{func_str}({all_args})"
    else:
        try:
            if hasattr(ast, 'unparse'): return ast.unparse(node)
            return f"<{type(node).__name__}>"
        except Exception: return f"<{type(node).__name__}>"

def create_location_info(node: ast.AST, current_filepath: str) -> Optional[Dict[str, Any]]:
    if not current_filepath or not hasattr(node, 'lineno'): return None
    end_lineno = getattr(node, 'end_lineno', node.lineno)
    end_col_offset = getattr(node, 'end_col_offset', -1)
    col_offset = getattr(node, 'col_offset', -1) 
    return {
        "file": current_filepath,
        "line": node.lineno, "col": col_offset,
        "end_line": end_lineno, "end_col": end_col_offset,
    }

class ToolDefinitionExtractor(ast.NodeVisitor):
    DECORATOR_IDENTS = {"function_tool", "openai.beta.tools.function_tool"}
    CONSTRUCTOR_IDENT = "FunctionTool" 

    def __init__(self, filepath: str) -> None:
        super().__init__()
        self.current_filepath = filepath
        self.discovered_tools: dict[str, ToolDefinition] = {}
        self.tool_locations: dict[str, Dict] = {}

    def visit_FunctionDef(self, node): self._process_function(node)
    def visit_AsyncFunctionDef(self, node): self._process_function(node)

    def visit_Assign(self, node: ast.Assign):
        if isinstance(node.value, ast.Call) and is_matching_call(node.value, {self.CONSTRUCTOR_IDENT}):
            call_node = node.value
            tool_name = extract_kwarg_string(call_node, "name")
            if not tool_name:
                print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Found '{self.CONSTRUCTOR_IDENT}' assignment without 'name' string arg. Skipping.")
                return

            description = extract_kwarg_string(call_node, "description") or ""
            location = create_location_info(node, self.current_filepath)
            tool_def = ToolDefinition(name=tool_name, custom=True, description=description)

            for target in node.targets:
                target_name = get_identifier_string(target)
                if target_name:
                    self.discovered_tools[target_name] = tool_def
                    if location: self.tool_locations[target_name] = location
                else:
                     print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Cannot assign '{self.CONSTRUCTOR_IDENT}' to complex target: {represent_node(target)}")
        self.generic_visit(node)

    def _process_function(self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]):
        decorator_node = find_decorator_node(node, self.DECORATOR_IDENTS)
        if not decorator_node:
            self.generic_visit(node)
            return

        func_name = node.name
        tool_name = func_name 
        description = ast.get_docstring(node) or ""
        location = create_location_info(node, self.current_filepath)

        if isinstance(decorator_node, ast.Call):
            name_override = extract_kwarg_string(decorator_node, "name_override")
            if name_override: tool_name = name_override
            desc_override = extract_kwarg_string(decorator_node, "description_override")
            if desc_override: description = desc_override

        tool_def = ToolDefinition(name=tool_name, custom=True, description=description)
        self.discovered_tools[func_name] = tool_def
        if location:
             self.tool_locations[tool_name] = location
             if tool_name != func_name:
                 self.tool_locations[func_name] = location
        self.generic_visit(node)

def gather_tool_definitions(root_path: str) -> Tuple[dict[str, ToolDefinition], dict[str, Dict]]:
    all_tools: dict[str, ToolDefinition] = {}
    all_locations: dict[str, Dict] = {}
    for file_path in Path(root_path).rglob("*.py"):
        filepath_str = str(file_path)
        try:
            with open(file_path, "r", encoding='utf-8') as f: content = f.read()
            tree = ast.parse(content, filename=filepath_str)
            extractor = ToolDefinitionExtractor(filepath_str)
            extractor.visit(tree)
            all_tools.update(extractor.discovered_tools)
            all_locations.update(extractor.tool_locations)
        except SyntaxError as e: print(f"Warning: Skipping file {filepath_str} due to SyntaxError: {e}")
        except Exception as e: print(f"Warning: Skipping file {filepath_str} due to unexpected error: {e}") 

    print(f"Found {len(all_tools)} custom tool definitions/references.")
    return all_tools, all_locations

def load_existing_tool_defs(json_filename: str = "existing_tools.json") -> dict[str, ToolDefinition]:
    existing_tools: dict[str, ToolDefinition] = {}
    try:
        script_dir = Path(__file__).parent
        full_path = script_dir / json_filename
        if not full_path.exists():
             full_path = Path(json_filename)
             if not full_path.exists():
                  print(f"Info: Existing tools file not found at '{script_dir / json_filename}' or '{Path.cwd() / json_filename}'.")
                  return {}

        print(f"Loading existing tools from: {full_path}")
        with open(full_path, 'r', encoding='utf-8') as f: tool_data = json.load(f)
        if not isinstance(tool_data, dict): raise ValueError("JSON must contain an object")

        for tool_id, description in tool_data.items():
            if not isinstance(tool_id, str) or not isinstance(description, str):
                print(f"Warning: Skipping invalid entry in existing tools (key/value not strings): {tool_id}")
                continue
            existing_tools[tool_id] = ToolDefinition(name=tool_id, custom=False, description=description)
        print(f"Loaded {len(existing_tools)} existing tools.")
        return existing_tools
    except Exception as e:
        print(f"Error loading existing tools from '{json_filename}': {e}")
        return {}

class AgentDefinitionExtractor(ast.NodeVisitor):
    AGENT_IDENTS = {"Agent", "openai.beta.agents.Agent"}
    HANDOFF_HELPER_IDENT: Optional[str] = "handoff" 

    def __init__(
        self,
        filepath: str,
        available_tools: dict[str, ToolDefinition],
        accumulated_agents: dict[str, AgentDefinition], 
        var_agent_link: dict[str, str] 
    ):
        super().__init__()
        self.current_filepath = filepath
        self.available_tools = available_tools
        self.parsed_agents = accumulated_agents
        self.var_agent_mapping = var_agent_link
        self.local_handoff_vars: dict[str, str] = {} 

    def visit_Assign(self, node: ast.Assign):
        if isinstance(node.value, ast.Call):
            call_node = node.value
            if is_matching_call(call_node, self.AGENT_IDENTS):
                self._process_agent_assignment(node)
            elif self.HANDOFF_HELPER_IDENT and is_matching_call(call_node, {self.HANDOFF_HELPER_IDENT}):
                 self._process_handoff_var_assignment(node)
        elif isinstance(node.value, ast.Name) and len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
             source_var = node.value.id
             target_var = node.targets[0].id
             if source_var in self.var_agent_mapping:
                  self.var_agent_mapping[target_var] = self.var_agent_mapping[source_var]
        self.generic_visit(node)

    def _process_agent_assignment(self, node: ast.Assign):
        assert isinstance(node.value, ast.Call)
        call_node = node.value
        location = create_location_info(node, self.current_filepath)

        try:
            agent_def = self._parse_agent_call(call_node, location)
            if not agent_def: return
        except InvalidAgentDefinitionError as e:
            print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Invalid Agent constructor: {e}. Skipping.")
            return
        except Exception as e:
             print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Unexpected error parsing agent definition: {e}. Skipping.")
             return

        for target in node.targets:
            target_var_name = get_identifier_string(target) 
            if target_var_name:
                agent_def.variable_name = target_var_name
                agent_id = agent_def.name
                if agent_id in self.parsed_agents:
                     existing_var = self.parsed_agents[agent_id].variable_name
                     if existing_var and existing_var != target_var_name:
                          print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Agent name '{agent_id}' already linked to var '{existing_var}'. Re-linking to '{target_var_name}'.")
                self.parsed_agents[agent_id] = agent_def
                self.var_agent_mapping[target_var_name] = agent_id
            else:
                 print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Cannot assign Agent to complex target: {represent_node(target)}")

    def _process_handoff_var_assignment(self, node: ast.Assign):
        if not self.HANDOFF_HELPER_IDENT: return
        try:
            target_agent_ref = self._parse_handoff_call_target(node.value)
        except InvalidHandoffSpecError as e:
            print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Invalid handoff definition: {e}. Skipping.")
            return

        for target in node.targets:
            target_var_name = get_identifier_string(target)
            if target_var_name:
                self.local_handoff_vars[target_var_name] = target_agent_ref
            else:
                 print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Cannot assign handoff result to complex target: {represent_node(target)}")

    def _parse_agent_call(self, agent_call: ast.Call, location: Optional[Dict]) -> Optional[AgentDefinition]:
        try:
            name = extract_kwarg_string(agent_call, "name")
            if not name:
                 first_arg = get_positional_arg_node(agent_call, 0)
                 name = extract_string_literal(first_arg)
                 if not name: raise ValueError("Agent missing 'name' string arg")

            tools_list = get_kwarg_node(agent_call, "tools")
            tools: List[ToolDefinition] = []
            if isinstance(tools_list, ast.List):
                tools = self._identify_tools(tools_list.elts)

            handoffs_list = get_kwarg_node(agent_call, "handoffs")
            handoffs: List[str] = []
            if isinstance(handoffs_list, ast.List):
                 handoffs = self._identify_handoffs(handoffs_list.elts)

            model_ref: Optional[str] = None
            model_node = get_kwarg_node(agent_call, "model")
            if model_node:
                model_name_literal = extract_string_literal(model_node)
                if model_name_literal:
                    model_ref = model_name_literal
                else:
                    model_ref = represent_node(model_node)

            return AgentDefinition(name=name, tools=tools, handoffs=handoffs, source_location=location , model_ref=model_ref)
        except (ValueError, TypeError, ValidationError, ToolDefinitionNotFoundError) as e:
            raise InvalidAgentDefinitionError(f"Failed extraction for '{represent_node(agent_call)}': {e}") from e

    def _identify_tools(self, tool_nodes: List[ast.AST]) -> List[ToolDefinition]:
        identified_tools = []
        for node in tool_nodes:
            tool_ref = get_identifier_string(node) 
            if tool_ref:
                if tool_ref in self.available_tools:
                    identified_tools.append(self.available_tools[tool_ref])
                else:
                     print(f"Warning [{self.current_filepath}]: Unknown tool reference '{tool_ref}'. Skipping.")
            else:
                print(f"Warning [{self.current_filepath}]: Unsupported item in tools list: {represent_node(node)}. Skipping.")
        return identified_tools

    def _identify_handoffs(self, handoff_nodes: List[ast.AST]) -> List[str]:
         identified_targets = []
         for node in handoff_nodes:
             target_ref: Optional[str] = None
             agent_ref = get_identifier_string(node)
             if agent_ref:
                  target_ref = agent_ref 
             elif self.HANDOFF_HELPER_IDENT and isinstance(node, ast.Call) and is_matching_call(node, {self.HANDOFF_HELPER_IDENT}):
                 try:
                     target_ref = self._parse_handoff_call_target(node)
                 except InvalidHandoffSpecError as e:
                      print(f"Warning [{self.current_filepath}]: Invalid handoff() in list: {e}. Skipping.")

             if target_ref:
                 if target_ref in self.local_handoff_vars:
                      identified_targets.append(self.local_handoff_vars[target_ref])
                 else:
                      identified_targets.append(target_ref)
             else:
                 print(f"Warning [{self.current_filepath}]: Unsupported item in handoffs list: {represent_node(node)}. Skipping.")
         return identified_targets

    def _parse_handoff_call_target(self, handoff_call: ast.Call) -> str:
        if not self.HANDOFF_HELPER_IDENT: raise InvalidHandoffSpecError("Handoff helper not configured.")
        target_node = get_kwarg_node(handoff_call, "agent")
        if not target_node: target_node = get_positional_arg_node(handoff_call, 0)
        if not target_node: raise InvalidHandoffSpecError("handoff() call missing target agent.")
        target_name = get_identifier_string(target_node) 
        if not target_name: raise InvalidHandoffSpecError(f"handoff() target not simple identifier: {represent_node(target_node)}")
        return target_name 

def gather_agent_definitions(root_path: str, available_tools: dict[str, ToolDefinition]) -> Tuple[dict[str, AgentDefinition], dict[str, str]]:
    all_agents: dict[str, AgentDefinition] = {} 
    var_agent_link: dict[str, str] = {} 
    for file_path in Path(root_path).rglob("*.py"):
        filepath_str = str(file_path)
        try:
            with open(file_path, "r", encoding='utf-8') as f: content = f.read()
            tree = ast.parse(content, filename=filepath_str)
            extractor = AgentDefinitionExtractor(filepath_str, available_tools, all_agents, var_agent_link)
            extractor.visit(tree)
        except SyntaxError as e: print(f"Warning: Skipping file {filepath_str} due to SyntaxError: {e}")
        except Exception as e: print(f"Warning: Skipping file {filepath_str} due to unexpected error: {e}") 

    print(f"Found {len(all_agents)} agent definitions.")
    return all_agents, var_agent_link

class ExecutionStartFinder(ast.NodeVisitor):
    RUNNER_IDENTS = {"Runner", "openai.beta.agents.Runner"} 
    METHOD_IDENT = "run"

    def __init__(self, filepath: str):
        self.current_filepath = filepath
        self.start_points: List[Tuple[str, Dict]] = [] 

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, ast.Attribute):
            attr_node = node.func
            method_name = attr_node.attr
            base_node = attr_node.value 
            base_name = get_identifier_string(base_node) 
            is_run_call = (method_name == self.METHOD_IDENT and base_name in self.RUNNER_IDENTS)

            if is_run_call:
                location = create_location_info(node, self.current_filepath)
                if node.args:
                    agent_arg = node.args[0]
                    agent_ref = get_identifier_string(agent_arg) 
                    if agent_ref and location:
                        self.start_points.append((agent_ref, location))
                    else:
                         agent_ref_str = represent_node(agent_arg)
                         print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Found {base_name}.{method_name} call unable to resolve agent arg '{agent_ref_str}' to simple name.")
                         if agent_ref_str and location: 
                              self.start_points.append((agent_ref_str, location))
                else:
                     print(f"Warning [{self.current_filepath}:{getattr(node, 'lineno', '?')}]: Found {base_name}.{method_name} call with no arguments.")
        self.generic_visit(node)

def find_execution_starts(root_path: str) -> List[Tuple[str, Dict]]:
     all_starts = []
     for file_path in Path(root_path).rglob("*.py"):
        filepath_str = str(file_path)
        try:
            with open(file_path, "r", encoding='utf-8') as f: content = f.read()
            tree = ast.parse(content, filename=filepath_str)
            finder = ExecutionStartFinder(filepath_str)
            finder.visit(tree)
            all_starts.extend(finder.start_points)
        except SyntaxError as e: print(f"Warning: Skipping file {filepath_str} due to SyntaxError: {e}")
        except Exception as e: print(f"Warning: Skipping file {filepath_str} due to unexpected error: {e}")
     return all_starts

def build_graph_json(
    agents: dict[str, AgentDefinition], 
    custom_tools: dict[str, ToolDefinition], 
    custom_tool_locs: dict[str, Dict], 
    existing_tools: dict[str, ToolDefinition], 
    start_points: List[Tuple[str, Dict]], 
    var_agent_link: Dict[str, str] 
    ) -> Dict[str, List[Dict]]:

    nodes: List[Dict] = []
    edges: List[Dict] = []
    processed_ids: Set[str] = set()

    for agent_id, agent_def in agents.items():
        if agent_id in processed_ids: continue
        node_meta = {
            "variable_name": agent_def.variable_name,
            "tool_names": [t.name for t in agent_def.tools],
            "handoff_refs": agent_def.handoffs, 
            "model": agent_def.model_ref
        }
        node_meta = {k: v for k, v in node_meta.items() if v is not None}
        node = {
            "id": agent_def.name, "name": agent_def.name,
            "node_type": NodeType.AGENT.value,
            "source_location": agent_def.source_location,
            "metadata": node_meta 
        }
        nodes.append(node)
        processed_ids.add(agent_id)

    agent_tool_names = set()
    for agent_def in agents.values():
        for tool in agent_def.tools:
            agent_tool_names.add(tool.name) 

    for tool_id in agent_tool_names:
        if tool_id in processed_ids: continue 
        tool_def = custom_tools.get(tool_id)
        is_custom = True
        if not tool_def:
            tool_def = existing_tools.get(tool_id)
            is_custom = False
        if not tool_def:
             print(f"Warning: Tool '{tool_id}' referenced but definition missing. Skipping node.")
             continue 

        node_type = NodeType.CUSTOM_TOOL if is_custom else NodeType.PREDEFINED_TOOL
        location = custom_tool_locs.get(tool_id) if is_custom else None
        tool_node = {
            "id": tool_id, "name": tool_id,
            "function_name": tool_id if is_custom else None,
            "docstring": tool_def.description,
            "node_type": node_type.value,
            "source_location": location,
            "metadata": {"custom": is_custom}
        }
        nodes.append(tool_node)
        processed_ids.add(tool_id)

    for agent_id, agent_def in agents.items():
        agent_loc = agent_def.source_location
        for tool_def in agent_def.tools:
            tool_name = tool_def.name
            if tool_name in processed_ids: 
                edges.append({
                    "source": agent_id, "target": tool_name,
                    "edge_type": "tool_usage", "condition": {},
                    "metadata": {"definition_location": agent_loc}
                })

    for agent_id, agent_def in agents.items():
        agent_loc = agent_def.source_location
        for handoff_ref in agent_def.handoffs:
            target_id = var_agent_link.get(handoff_ref) 
            if not target_id: target_id = handoff_ref
            if target_id in agents: 
                edges.append({
                    "source": agent_id, "target": target_id,
                    "edge_type": "handoff", "condition": {},
                    "metadata": {"definition_location": agent_loc, "handoff_ref": handoff_ref}
                })
            else:
                 print(f"Warning: Agent '{agent_id}' handoff ref '{handoff_ref}' (->'{target_id}') not found.")

    start_id, end_id = "Start", "End"
    if start_id not in processed_ids: nodes.append({"id": "Start", "name": "Start", "node_type": NodeType.START.value}); processed_ids.add(start_id)
    if end_id not in processed_ids: nodes.append({"id": "End", "name": "End", "node_type": NodeType.END.value}); processed_ids.add(end_id)

    runner_starts = set()
    for agent_ref, loc in start_points:
        agent_id = var_agent_link.get(agent_ref) 
        if agent_id and agent_id in agents:
            if not any(e['source'] == start_id and e['target'] == agent_id for e in edges): 
                edges.append({
                    "source": start_id, "target": agent_id,
                    "edge_type": "execution_start", "condition": {},
                    "metadata": {"definition_location": loc, "runner_agent_ref": agent_ref}
                })
            runner_starts.add(agent_id)
        else:
            print(f"Warning: Runner start ref '{agent_ref}' not mapped to known agent.")

    agent_ids = set(agents.keys())
    incoming: Dict[str, int] = {name: 0 for name in agent_ids}
    outgoing: Dict[str, int] = {name: 0 for name in agent_ids}
    for edge in edges: 
        if edge['edge_type'] == 'handoff' and edge['source'] in agent_ids and edge['target'] in agent_ids:
            incoming[edge['target']] = incoming.get(edge['target'], 0) + 1
            outgoing[edge['source']] = outgoing.get(edge['source'], 0) + 1

    for agent_id in agent_ids:
        if incoming.get(agent_id, 0) == 0 and agent_id not in runner_starts:
            if not any(e['source'] == start_id and e['target'] == agent_id for e in edges):
                edges.append({"source": start_id, "target": agent_id, "edge_type": "implicit_start", "condition": {}, "metadata": {}})
    for agent_id in agent_ids:
        if outgoing.get(agent_id, 0) == 0:
             if not any(e['source'] == agent_id and e['target'] == end_id for e in edges):
                 edges.append({"source": agent_id, "target": end_id, "edge_type": "implicit_end", "condition": {}, "metadata": {}})

    for node in nodes:
        node.setdefault("function_name", None)
        node.setdefault("docstring", None)
        node.setdefault("source_location", None)
        node.setdefault("metadata", {})

    for node in nodes:
        if "metadata" in node:
             node["metadata"] = {k:v for k,v in node["metadata"].items() if v is not None}


    return {"nodes": nodes, "edges": edges}



def extract_openai_agents_graph(scan_path: str, output_file: str):
    existing_tools_file = "existing_tools.json" # Hardcoded path

    if not os.path.isdir(scan_path):
        print(f"Error: Path '{scan_path}' is not a valid directory.")
        sys.exit(1)


    print(f"Analyzing agent structures in: {scan_path}")

    try:
        custom_tools, custom_locs = gather_tool_definitions(scan_path)
        existing_tools = load_existing_tool_defs(existing_tools_file)
        available_tools = {**existing_tools, **custom_tools}
        print(f"Total available tools: {len(available_tools)}")

        agents, var_agent_link = gather_agent_definitions(scan_path, available_tools)

        start_points = find_execution_starts(scan_path)
        print(f"Found {len(start_points)} potential execution starts.")

        final_graph = build_graph_json(
            agents=agents,
            custom_tools=custom_tools,
            custom_tool_locs=custom_locs,
            existing_tools=existing_tools,
            start_points=start_points,
            var_agent_link=var_agent_link
        )
        print(f"Analysis complete. Nodes: {len(final_graph['nodes'])}, Edges: {len(final_graph['edges'])}.")

    except Exception as e:
         print(f"Error during graph extraction: {e}")
         sys.exit(1)


    if final_graph["nodes"] or final_graph["edges"]:
        try:

            final_graph["nodes"].sort(key=lambda x: x.get('id', ''))
            final_graph["edges"].sort(key=lambda x: (x.get('source', ''), x.get('target', ''), x.get('edge_type', '')))


            output_path = Path(output_file)
            output_path.parent.mkdir(parents=True, exist_ok=True)

            with open(output_path, "w", encoding='utf-8') as f:
                json.dump(final_graph, f, indent=2)
            print(f"\nGraph structure written to {output_path}")


            print("\n--- Summary ---")
            if not final_graph["nodes"]: print("No nodes found.")
            for node in final_graph["nodes"]:
                 loc = node.get('source_location')
                 loc_str = f" ({loc['file']}:{loc['line']})" if loc and loc.get('file') else ""
                 node_id = node.get('id', 'Unknown ID')
                 node_type = node.get('node_type', 'Unknown Type')
                 print(f"Node: {node_id} ({node_type}){loc_str}")

            if not final_graph["edges"]: print("No edges found.")
            for edge in final_graph["edges"]:
                 cond_str = f" Cond: {edge.get('condition')}" if edge.get('condition') else ""
                 meta_loc = edge.get('metadata', {}).get('definition_location')
                 loc_str = f" ({meta_loc['file']}:{meta_loc['line']})" if meta_loc and meta_loc.get('file') else ""
                 source = edge.get('source', 'Unknown')
                 target = edge.get('target', 'Unknown')
                 edge_type = edge.get('edge_type', 'Unknown Type')
                 print(f"Edge: {source} -> {target} ({edge_type}){cond_str}{loc_str}")

        except Exception as e:
            print(f"Error writing JSON output or printing summary: {e}")
            sys.exit(1)
    else:
        print("\nNo graph structure found or extracted.")
        try:

            output_path = Path(output_file)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            with open(output_path, "w", encoding='utf-8') as f:
                json.dump({"nodes": [], "edges": []}, f, indent=2)
            print(f"Empty graph structure written to {output_path}")
        except Exception as e:
            print(f"Error writing empty JSON output to {output_path}: {e}")
            sys.exit(1)



if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Parse Python code to extract OpenAI Agent structures into JSON.")
    parser.add_argument("directory", type=str, help="Directory containing the Python code.")
    parser.add_argument("-o", "--output", type=str, default="agent_graph_structure.json", help="Output JSON filename.")

    args = parser.parse_args()

    target_scan_dir = args.directory
    output_file = args.output


    extract_openai_agents_graph(target_scan_dir, output_file)